import logging
import os

import numpy as np

from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType

logger = logging.getLogger(__name__)


class SHBinaryWriter:
    def __init__(self, filename, options):
        self.filename = filename

    def write(self, estimator):
        # TODO
        pass


class TxtWriter:
    @staticmethod
    def _axis_name(geo_type, axis_no):
        cyl = ('R', 'PHI', 'Z')
        msh = ('X', 'Y', 'Z')
        if geo_type in (SHGeoType.cyl, SHGeoType.dcyl):
            return cyl[axis_no]
        return msh[axis_no]

    def __init__(self, filename, options):
        if filename.endswith(".txt"):
            self.filename = filename
        else:
            self.filename = filename + ".txt"
        self.ax = ''
        self.ay = ''
        self.az = ''

    @staticmethod
    def _header_first_line(estimator):
        """first line with estimator geo type"""
        result = "#   DETECTOR OUTPUT\n"
        if estimator.geotyp in (SHGeoType.plane, SHGeoType.dplane,):
            result = "#             DETECTOR OUTPUT PLANE/DPLANE\n"
        elif estimator.geotyp in (SHGeoType.zone, SHGeoType.dzone,):
            result = "#             DETECTOR OUTPUT ZONE/DZONE\n"
        elif estimator.geotyp in (SHGeoType.msh, SHGeoType.dmsh,):
            result = "#   DETECTOR OUTPUT MSH/DMSH\n"
        elif estimator.geotyp == SHGeoType.geomap:
            result = "#   DETECTOR OUTPUT GEOMAP\n"
        return result

    def _header_geometric_info(self, det):
        """next block - scoring object geometrical information"""

        from pymchelper.writers.fortranformatter import format_d
        result = ""
        if det.geotyp in {SHGeoType.plane, SHGeoType.dplane}:
            result += "#   PLANE point(X,Y,Z)         :"
            result += f"{format_d(10, 3, det.sx)}{format_d(10, 3, det.sy)}{format_d(10, 3, det.sz)}\n"
            result += "#   PLANE normal vect(Vx,Vy,Vz):"
            result += f"{format_d(10, 3, det.nx)}{format_d(10, 3, det.ny)}{format_d(10, 3, det.nz)}\n"
        elif det.geotyp in {SHGeoType.zone, SHGeoType.dzone}:
            result += f"#   ZONE START:{int(det.x.min_val):6d} ZONE END:{int(det.x.max_val):6d}\n"
        else:
            result += f"#   {self.ax} BIN:{det.x.n:6d} {self.ay} BIN:{det.y.n:6d} {self.az} BIN:{det.z.n:6d}\n"
        return result

    @staticmethod
    def _header_scored_value(geotyp, dettyp, particle):
        """scored value and optionally particle type"""
        result = ""
        if geotyp != SHGeoType.geomap and particle:
            result += f"#   JPART:{particle:6d} DETECTOR TYPE: {str(dettyp).ljust(10)}\n"
        else:
            det_type_name = str(dettyp)
            if dettyp in (SHDetType.zone, SHDetType.medium,):
                det_type_name += "#"
            result += f"#                DETECTOR TYPE: {str(det_type_name).ljust(10)}\n"
        return result

    def _header_no_of_bins_and_prim(self, estimator):
        from pymchelper.writers.fortranformatter import format_d

        header = ""
        # number of bins in each dimensions
        if estimator.geotyp not in (SHGeoType.plane, SHGeoType.dplane, SHGeoType.zone, SHGeoType.dzone):
            header += f"#   {self.ax} START:{format_d(10, 3, estimator.x.min_val)}"
            header += f" {self.ay} START:{format_d(10, 3, estimator.y.min_val)}"
            header += f" {self.az} START:{format_d(10, 3, estimator.z.min_val)}\n"
            header += f"#   {self.ax} END  :{format_d(10, 3, estimator.x.max_val)}"
            header += f" {self.ay} END  :{format_d(10, 3, estimator.y.max_val)}"
            header += f" {self.az} END  :{format_d(10, 3, estimator.z.max_val)}\n"

        # number of primaries
        header += f"#   PRIMARIES:{format_d(10, 3, estimator.number_of_primaries)}\n"

        return header

    def write(self, estimator):
        """TODO"""
        # save to single page to a file without number (i.e. output.dat)
        if len(estimator.pages) == 1:
            self.write_single_page(estimator.pages[0], self.filename)
        else:
            # split output path into directory, basename and extension
            dir_path = os.path.dirname(self.filename)
            if not os.path.exists(dir_path):
                logger.info("Creating: %s", dir_path)
                os.makedirs(dir_path)
            file_base_part, file_ext = os.path.splitext(os.path.basename(self.filename))

            # loop over all pages and save an image for each of them
            for i, page in enumerate(estimator.pages):

                # calculate output filename. it will include page number padded with zeros.
                # for 10-99 pages the filename would look like: output_p01.png, ... output_p99.png
                # for 100-999 pages the filename would look like: output_p001.png, ... output_p999.png
                zero_padded_page_no = str(i + 1).zfill(len(str(len(estimator.pages))))
                output_filename = f"{file_base_part}_p{zero_padded_page_no}{file_ext}"
                output_path = os.path.join(dir_path, output_filename)

                # save the output file
                logger.info("Writing: %s", output_path)
                self.write_single_page(page, output_path)

        return 0

    def write_single_page(self, page, filename):
        """TODO"""
        logger.info("Writing: %s", filename)

        from pymchelper.writers.fortranformatter import format_e

        self.ax = self._axis_name(page.estimator.geotyp, 0)
        self.ay = self._axis_name(page.estimator.geotyp, 1)
        self.az = self._axis_name(page.estimator.geotyp, 2)

        # original bdo2txt is not saving header data for some of cylindrical scorers, hence we do the same
        if page.estimator.geotyp in (SHGeoType.cyl, SHGeoType.dcyl,) and \
                page.dettyp in (SHDetType.fluence, SHDetType.avg_energy, SHDetType.avg_beta, SHDetType.energy):
            header = ""
        else:
            header = self._header_first_line(page.estimator)

            header += self._header_geometric_info(page.estimator)

            header += self._header_scored_value(
                page.estimator.geotyp, page.dettyp, getattr(page.estimator, 'particle', None))

            header += self._header_no_of_bins_and_prim(page.estimator)

        # dump data
        with open(filename, 'w') as fout:  # skipcq: PTC-W6004
            logger.info("Writing: %s", filename)
            fout.write(header)

            det_error = page.error_raw.ravel()
            if np.all(np.isnan(page.error_raw)):
                det_error = [None] * page.data_raw.size
            xmesh = page.axis(0)
            ymesh = page.axis(1)
            zmesh = page.axis(2)

            logger.debug('xmesh %s', str(xmesh))
            logger.debug('ymesh %s', str(ymesh))
            logger.debug('zmesh %s', str(zmesh))

            zlist, ylist, xlist = np.meshgrid(zmesh.data, ymesh.data, xmesh.data, indexing='ij')

            logger.debug('xlist %s', str(xlist))
            logger.debug('ylist %s', str(ylist))
            logger.debug('zlist %s', str(zlist))

            for x, y, z, v, e in zip(xlist.ravel(), ylist.ravel(), zlist.ravel(), page.data.ravel(), det_error):
                if page.estimator.geotyp in {SHGeoType.zone, SHGeoType.dzone}:
                    x = 0.0
                # dirty hack to be compliant with old bdo2txt and files generated in old (<0.6) BDO format
                # this hack will be removed at some point together with bdo-style converter
                elif not hasattr(page.estimator, "mc_code_version") and page.estimator.geotyp == SHGeoType.plane:
                    x = (page.estimator.sx + page.estimator.nx) / 2.0
                    y = (page.estimator.sy + page.estimator.ny) / 2.0
                    z = (page.estimator.sz + page.estimator.nz) / 2.0
                else:
                    x = float('nan') if np.isnan(x) else x
                y = float('nan') if np.isnan(y) else y
                z = float('nan') if np.isnan(z) else z
                v = float('nan') if np.isnan(v) else v
                tmp = format_e(14, 7, x) + " " + format_e(14, 7, y) + " " + \
                    format_e(14, 7, z) + " " + format_e(23, 16, v)

                if e is not None:
                    e = float('nan') if np.isnan(e) else e
                    tmp += " " + format_e(23, 16, e)

                tmp += "\n"

                fout.write(tmp)

        return 0



from pymchelper.shieldhit.detector.estimator_type import SHGeoType
from pymchelper.shieldhit.detector.geometry import Zone, Plane
from pymchelper.shieldhit.particle import SHParticleType


class CardLine:
    """
    Represents single line of detect.dat file
    """

    comment = "*----0---><----1---><----2---><----3---><----4---><----5---><----6--->"
    credits = "* generated by pymchelper (https://github.com/DataMedSci/pymchelper) *"
    no_of_elements = 7
    element_length = 10

    def __init__(self, data=None):
        self.data = []
        if data is None:
            self.data = [" " * CardLine.element_length] * CardLine.no_of_elements
        else:
            for element in data:
                if isinstance(element, SHGeoType):
                    elem_string = CardLine.any_to_element(element, align_right=False)
                else:
                    elem_string = CardLine.any_to_element(element)
                self.data.append(elem_string)

    def __str__(self):
        return "".join(self.data)

    @staticmethod
    def any_to_element(s, align_right=True):
        if len(str(s)) > CardLine.element_length:
            raise Exception("Element [{:s}] should have {:d}, not {:d} elements".format(
                str(s), CardLine.element_length, len(str(s))))
        if s is None:
            s = ""
        padding = " " * (CardLine.element_length - len(str(s)))
        result = padding + str(s)
        if not align_right:
            result = str(s) + padding
        return result


class EstimatorWriter:
    """
    Helper class. Gives textual representation of Estimator objects.
    """

    @staticmethod
    def get_lines(estimator):
        """
        Converts estimator to CardLine objects
        :param estimator: valid estimator object
        :return: tuple of CardLine objects
        """
        if estimator.particle_type == SHParticleType.heavy_ion:
            data_heavy_ion = ["", estimator.heavy_ion_type.z, estimator.heavy_ion_type.a, "", "", "", ""]
        if isinstance(estimator.geometry, Zone):
            data1 = [
                estimator.estimator, estimator.geometry.start, estimator.geometry.stop, "",
                estimator.particle_type.value, estimator.detector_type, estimator.filename
            ]
            if estimator.particle_type == SHParticleType.heavy_ion:
                return CardLine(data1), CardLine(data_heavy_ion)
            return (CardLine(data1),)
        if isinstance(estimator.geometry, Plane):
            data1 = [
                estimator.estimator, estimator.geometry.point_x, estimator.geometry.point_y, estimator.geometry.point_z,
                estimator.geometry.normal_x, estimator.geometry.normal_y, estimator.geometry.normal_z
            ]
            data2 = ["", "", "", "", estimator.particle_type.value, estimator.detector_type, estimator.filename]
            if estimator.particle_type == SHParticleType.heavy_ion:
                return CardLine(data1), CardLine(data2), CardLine(data_heavy_ion)
            return CardLine(data1), CardLine(data2)
        data1 = [
            estimator.estimator, estimator.geometry.axis[0].start, estimator.geometry.axis[1].start,
            estimator.geometry.axis[2].start, estimator.geometry.axis[0].stop, estimator.geometry.axis[1].stop,
            estimator.geometry.axis[2].stop
        ]
        data2 = ["", estimator.geometry.axis[0].nbins, estimator.geometry.axis[1].nbins,
                 estimator.geometry.axis[2].nbins, estimator.particle_type.value, estimator.detector_type,
                 estimator.filename]
        if estimator.particle_type == SHParticleType.heavy_ion:
            return CardLine(data1), CardLine(data2), CardLine(data_heavy_ion)
        return CardLine(data1), CardLine(data2)

    @staticmethod
    def get_text(estimator, add_comment=False):
        """
        Converts Estimator to text
        :param estimator: valid Estimator object
        :param add_comment: if True, prepends textual representation with comment line
        :return: text representation (set of text lines) of Estimator
        """
        code_lines = EstimatorWriter.get_lines(estimator)
        result = ""
        if add_comment:
            result = CardLine.comment + "\n"
        result += "\n".join(map(str, code_lines)) + "\n"
        return result


class EstimatorReader:
    pass
    # TODO to be implemented



from copy import deepcopy
import logging
import os
from pathlib import Path
import re
import shutil
import subprocess
import timeit
from multiprocessing import Pool

from enum import IntEnum
from typing import Tuple
from pymchelper.flair import Input
from pymchelper.executor.options import SimulationSettings

from pymchelper.simulator_type import SimulatorType
from pymchelper.input_output import frompattern, get_topas_estimators


class OutputDataType(IntEnum):
    """
    Output type requested by user plots (i.e .png) or text data
    """
    plot = 1
    txt = 2


class Runner:
    """
    Main class responsible for configuring and starting multiple parallel MC simulation processes
    It can be used to access combined averaged results of the simulation.
    """
    def __init__(self, settings: SimulationSettings, jobs: int=None,
                 keep_workspace_after_run: bool=False, output_directory: str='.'):
        self.settings = settings

        # create pool of processes, waiting to be started by run method
        # if jobs is not specified, os.cpu_count() would be used
        self._pool = Pool(processes=jobs)

        # User of the runner has two options: either to specify number of parallel jobs by
        # setting the self.jobs to given number, or to leave it as None. If self.jobs is None
        # then multiprocessing library will automatically allocate number of parallel jobs to `os.cpu_count()`
        # Therefore we cannot rely of self.jobs as the actual counter of parallel processes
        # Instead we use undocumented feature of multiprocessing module,
        # extracting actual number of allocated processes from `_processes` attribute in Pool class
        # To see how it is used internally in the Python (in v3.9) source code take a look at:
        # https://github.com/python/cpython/blob/3.9/Lib/multiprocessing/pool.py#L210
        self.jobs = self._pool._processes

        # workspace is a collection of working directories
        # this manager is responsible for creating and cleaning working directories
        self.workspace_manager = WorkspaceManager(output_directory=output_directory,
                                                  keep_workspace_after_run=keep_workspace_after_run)

    def run(self):
        """
        Execute parallel simulation processes, creating workspace (and working directories) in the `output_directory`
        In case of successful execution return True, otherwise return False
        """
        start_time = timeit.default_timer()

        if self.settings.simulator_type in [SimulatorType.shieldhit, SimulatorType.fluka]:
            # SHIELD-HIT12A and Fluka require RNG seeds to be integers greater or equal to 1
            # each of the workers needs to have its own different RNG seed
            rng_seeds = range(1, self.jobs + 1)

        elif self.settings.simulator_type == SimulatorType.topas:
            # For TOPAS we don't need to create multiple working directories and a pool of workers,
            # as we can use embedded parallelization in TOPAS.
            # For that we need to modify the input file and set the number of threads to the number of jobs.
            # We set one rng seed, so one working directory and one worker will be created.

            modified_input_filename = Path(self.settings.input_path).name.replace(".txt", "_modified.txt")
            modified_input_path = Path(self.settings.input_path).parent / modified_input_filename

            with open(self.settings.input_path, 'r') as f:
                config = f.read()
                if "i:Ts/NumberOfThreads" in config:
                    pattern = r"i:Ts/NumberOfThreads\s*=\s*\d+"
                    replacement = f"i:Ts/NumberOfThreads = {self.jobs}"
                    config = re.sub(pattern, replacement, config)
                else:
                    config = f"i:Ts/NumberOfThreads = {self.jobs}\n" + config

            modified_input_path.write_text(config)

            self.settings.input_path = str(modified_input_path)

            rng_seeds = [1]

        # create working directories
        self.workspace_manager.create_working_directories(simulation_input_path=self.settings.input_path,
                                                          rng_seeds=rng_seeds)

        if self.settings.simulator_type == SimulatorType.fluka:
            for seed, workdir in zip(rng_seeds, self.workspace_manager.working_directories_abs_paths):
                destination = str(Path(workdir) / Path(self.settings.input_path).name)
                self.__update_fluka_input_file(destination, seed)

        # rng seeds injection to settings for each SingleSimulationExecutor call
        # TODO consider better way of doing it  # skipcq: PYL-W0511
        settings_list = []
        for rng_seed in rng_seeds:
            current_settings = deepcopy(self.settings)  # do not modify original arguments
            if self.settings.simulator_type in [SimulatorType.shieldhit, SimulatorType.topas]:
                current_settings.set_rng_seed(rng_seed)
            settings_list.append(current_settings)

        # create executor callable object for current run
        executor = SingleSimulationExecutor()

        try:
            # start execution using pool of workers, mapping the executor callable object to the different settings
            # and workspaces
            self._pool.map(executor, zip(settings_list, self.workspace_manager.working_directories_abs_paths))
        except KeyboardInterrupt:
            logging.info('Terminating the pool')
            self._pool.terminate()
            logging.info('Pool is terminated')
            self.workspace_manager.clean()
            return False

        elapsed = timeit.default_timer() - start_time
        logging.info("run elapsed time {:.3f} seconds".format(elapsed))

        return True

    def get_data(self):
        """
        Scans the output directory for location of the working directories (like run_1, run_2).
        Takes all files from all working directories in `output_dir`,
        merges their content to form pymchelper Estimator objects.
        For each of the output file a single Estimator objects is created, which holds numpy arrays with results.
        Return dictionary with keys being output filenames, and values being Estimator objects
        """
        start_time = timeit.default_timer()

        # scans output directory for MC simulation output files
        estimators_dict = {}
        estimators_list = []

        if self.settings.simulator_type == SimulatorType.shieldhit:
            output_files_pattern = str(Path(self.workspace_manager.output_dir_absolute_path) / "run_*" / "*.bdo")
            logging.debug("Files to merge %s", output_files_pattern)
            # convert output files to list of estimator objects
            estimators_list = frompattern(output_files_pattern)

        elif self.settings.simulator_type == SimulatorType.topas:
            output_files_path = str(Path(self.workspace_manager.output_dir_absolute_path) / "run_1")
            estimators_list = get_topas_estimators(output_files_path)

        elif self.settings.simulator_type == SimulatorType.fluka:
            output_files_pattern = os.path.join(self.workspace_manager.output_dir_absolute_path, "run_*", "*_fort.*")
            logging.debug("Files to merge %s", output_files_pattern)
            estimators_list = frompattern(output_files_pattern)

        for estimator in estimators_list:
            logging.debug("Appending estimator for {:s}".format(estimator.file_corename))
            estimators_dict[estimator.file_corename] = estimator
        elapsed = timeit.default_timer() - start_time
        logging.info("Workspace reading {:.3f} seconds".format(elapsed))

        return estimators_dict

    def clean(self):
        """Removes all working directories (if exists)"""
        self.workspace_manager.clean()

    @staticmethod
    def __update_fluka_input_file(destination: str, rng_seed: int):
        """Updates the FLUKA input file with the new RNG seed."""
        with open(destination, 'r') as destination_file:
            lines = destination_file.readlines()
        has_randomize_card = False
        for index, line in enumerate(lines):
            if line.startswith('RANDOMIZ'):
                card = Input.Card("RANDOMIZ")
                card.setWhat(1, 1.)
                card.setWhat(2, rng_seed)
                lines[index] = card.toStr() + '\n'
                has_randomize_card = True
        if not has_randomize_card:
            card = Input.Card("RANDOMIZ")
            card.setWhat(1, 1.)
            card.setWhat(2, rng_seed)
            lines.append(card.toStr() + '\n')

        with open(destination, 'w') as destination_file:
            destination_file.writelines(lines)


class SingleSimulationExecutor:
    """
    Callable class responsible for execution of the single MC simulation process.
    """

    def __call__(self, settings_and_working_dir: Tuple[SimulationSettings, str], **kwargs):

        # we deliberately combine settings and list of working directories
        # in the single argument `settings_and_working_dir`
        # as this would simplify using this class by multiprocessing module
        settings, working_dir_abs_path = settings_and_working_dir
        try:
            # combine MC engine executable with its command line options to form core of the command string
            # this will form basis of the command, like:
            # /usr/local/bin/shieldhit --time 00:30:50 -v -N 3
            core_command_string = str(settings)

            # for easier digesting by subprocess module, convert command string to a list
            # and finally append the location of the input files
            # finally we obtain a list like:
            # ('/usr/local/bin/shieldhit', '--time', '00:30:50', '-v', '-N', '3', '/data/my/simulation/input')
            command_as_list = core_command_string.split()
            if settings.simulator_type == SimulatorType.fluka:
                input_file_path = str(Path(working_dir_abs_path) / Path(settings.input_path).name)
                command_as_list.append(input_file_path)
            else:
                command_as_list.append(working_dir_abs_path)

            # execute the MC simulation on a spawned process
            # TODO handle this differently, i.e. redirect it to file or save in some variable   # skipcq: PYL-W0511
            logging.debug('working directory {:s}, command {:s}'.format(working_dir_abs_path,
                                                                        ' '.join(command_as_list)))
            DEVNULL = open(os.devnull, 'wb')
            subprocess.check_call(command_as_list, cwd=working_dir_abs_path, stdout=DEVNULL, stderr=DEVNULL)
        except KeyboardInterrupt:
            logging.debug("KeyboardInterrupt")
            raise KeyboardInterrupt


class WorkspaceManager:
    """
    A workspace consists of multiple working directories (i.e. run_1, run_2),
    each per one of the parallel simulation run.
    """
    def __init__(self, output_directory='.', keep_workspace_after_run=False):
        self.output_dir_absolute_path = os.path.abspath(output_directory)
        self.keep_workspace_after_run = keep_workspace_after_run
        self.working_directories_abs_paths = []

    def create_working_directories(self, simulation_input_path, rng_seeds=()):
        """
        Create working directories and fill `self.working_directories_abs_paths`
        """
        self.working_directories_abs_paths = []
        for rng_seed in rng_seeds:
            # set workspace to a subdirectory of the output_directory, with run_* pattern
            # i.e. /home/user/output/run_3
            working_dir_abs_path = os.path.join(self.output_dir_absolute_path, 'run_{:d}'.format(rng_seed))
            logging.info("Workspace {:s}".format(working_dir_abs_path))

            if os.path.isdir(simulation_input_path):
                # if path already exists, remove it before copying with copytree()
                if os.path.exists(working_dir_abs_path):
                    shutil.rmtree(working_dir_abs_path)
                # if cleaned or not existing, then create it
                if not os.path.exists(working_dir_abs_path):
                    os.makedirs(working_dir_abs_path)
                # copy all files from the directory
                for directory_entry in os.listdir(simulation_input_path):
                    path_to_directory_entry = os.path.join(simulation_input_path, directory_entry)
                    if os.path.isfile(path_to_directory_entry):
                        shutil.copy2(path_to_directory_entry, working_dir_abs_path)
                logging.debug("Copying input files into {:s}".format(working_dir_abs_path))
            elif os.path.isfile(simulation_input_path):
                if not os.path.exists(working_dir_abs_path):
                    os.makedirs(working_dir_abs_path)
                shutil.copy2(simulation_input_path, working_dir_abs_path)
                logging.debug("Copying input files into {:s}".format(working_dir_abs_path))
            else:
                logging.debug("Input files {:s} not a dir or file".format(simulation_input_path))

            self.working_directories_abs_paths.append(working_dir_abs_path)

    def clean(self):
        """
        clean the workspace by removing all working directories
        (only if requested by `keep_workspace_after_run` flag)
        """
        if not self.keep_workspace_after_run:
            start_time = timeit.default_timer()
            for working_dir_abs_path in self.working_directories_abs_paths:
                # shutil.rmtree will throw exception if the directory we are trying to remove doesn't exist
                # hence we only remove existing directories
                # this allows safely to call `clean` method multiple times
                if os.path.exists(working_dir_abs_path):
                    shutil.rmtree(working_dir_abs_path)
            elapsed = timeit.default_timer() - start_time
            print("Cleaning {:.3f} seconds".format(elapsed))



from abc import abstractmethod
import logging
from pathlib import Path
from pymchelper.estimator import Estimator

logger = logging.getLogger(__name__)


class Writer:
    """Base class for all writers."""

    def __init__(self, output_path: str):
        self.output_path = Path(output_path)

    def write(self, estimator: Estimator) -> int:
        """Write the estimator data to a file."""
        # create output directory if it does not exist
        self.output_path.parent.mkdir(parents=True, exist_ok=True)

        # save to single page to a file without number (i.e. output.dat)
        if len(estimator.pages) == 1:
            self.write_single_page(page=estimator.pages[0], output_path=self.output_path)
        else:
            # loop over all pages and save an image for each of them
            for i, page in enumerate(estimator.pages):

                # calculate output filename. it will include page number padded with zeros.
                # for 10-99 pages the filename would look like: output_p01.png, ... output_p99.png
                # for 100-999 pages the filename would look like: output_p001.png, ... output_p999.png
                zero_padded_page_no = str(i + 1).zfill(len(str(len(estimator.pages))))
                page_output_path = self.output_path.parent
                page_output_path /= f"{self.output_path.stem}_p{zero_padded_page_no}{self.output_path.suffix}"

                # save the output file
                logger.info("Writing %s", page_output_path)
                self.write_single_page(page=page, output_path=page_output_path)
        return 1

    @abstractmethod
    def write_single_page(self, page, output_path: Path):
        """Write a single page to a file."""
        raise NotImplementedError



from enum import IntEnum
from typing import List, Optional


class SimulatorType(IntEnum):
    """Type of the MC simulator: SHIELD-HIT12A, TOPAS or FLUKA"""

    shieldhit = 1
    topas = 2
    fluka = 3

    @classmethod
    @property
    def names(cls) -> List[str]:
        """Return list of available simulators"""
        return [s.name for s in SimulatorType]

    @staticmethod
    def from_name(name: str) -> Optional['SimulatorType']:
        """Return simulator type from its name"""
        try:
            return SimulatorType[name]
        except KeyError:
            return None



from abc import abstractmethod
import logging

import numpy as np

logger = logging.getLogger(__name__)


class ReaderFactory(object):
    def __init__(self, filename):
        self.filename = filename

    @abstractmethod
    def get_reader(self):
        pass


class Reader(object):
    def __init__(self, filename):
        self.filename = filename

    def read(self, estimator):
        result = self.read_data(estimator)
        if not result:
            return False
        for page in estimator.pages:
            page.error_raw = np.zeros_like(page.data_raw) * np.nan
        return True

    @abstractmethod
    def read_data(self, estimator):
        pass

    @property
    @abstractmethod
    def corename(self):
        pass



import logging
import os

import numpy as np

from pymchelper.axis import MeshAxis
from pymchelper.readers.common import Reader
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType

logger = logging.getLogger(__name__)


class SHReader(Reader):
    """
    Reads binary output files generated by SHIELD-HIT12A code.
    """

    def read_data(self, estimator, nscale=1):
        """
        TODO
        :param estimator:
        :param nscale:
        :return:
        """
        _postprocess(estimator, nscale)
        return True

    @property
    def corename(self):
        """
        TODO
        :return:
        """
        core_name = None

        if self.filename.endswith(('.bdo', '.bdox')):  # TODO add more sophisticated check for file being SH12A output
            basename = os.path.basename(self.filename)
            # we expect the basename to follow one of two conventions:
            #  - corenameABCD.bdo (where ABCD is 4-digit integer)
            #  - corename.bdo
            core_name = basename[:-4]  # assume no number in the basename
            if basename[-8:-4].isdigit() and len(basename[-8:-4]) == 4:  # check if number present
                core_name = basename[:-8]

        return core_name


def mesh_unit_and_name(estimator, axis):
    """ Gets units and names depending on detector type.
    """
    _geotyp_units = {
        SHGeoType.msh: ("cm", "cm", "cm"),
        SHGeoType.dmsh: ("cm", "cm", "cm"),
        SHGeoType.cyl: ("cm", "radians", "cm"),
        SHGeoType.dcyl: ("cm", "radians", "cm"),
        SHGeoType.zone: ("zone number", "(nil)", "(nil)"),
        SHGeoType.voxscore: ("cm", "cm", "cm"),
        SHGeoType.geomap: ("cm", "cm", "cm"),
        SHGeoType.plane: ("cm", "cm", "cm"),
        SHGeoType.dplane: ("cm", "cm", "cm")
    }
    _default_units = ("(nil)", "(nil)", "(nil)")

    unit = _geotyp_units.get(estimator.geotyp, _default_units)[axis]

    if estimator.geotyp in {SHGeoType.msh, SHGeoType.dmsh, SHGeoType.voxscore, SHGeoType.geomap,
                            SHGeoType.plane, SHGeoType.dplane}:
        name = ("Position (X)", "Position (Y)", "Position (Z)")[axis]
    elif estimator.geotyp in {SHGeoType.cyl, SHGeoType.dcyl}:
        name = ("Radius (R)", "Angle (PHI)", "Position (Z)")[axis]
    else:
        name = ""

    # dirty hack to change the units for differential scorers
    if hasattr(estimator, 'dif_axis') and hasattr(estimator, 'dif_type') and axis == estimator.dif_axis:
        if estimator.dif_type == 1:
            unit, name = _get_detector_unit(SHDetType.energy, estimator.geotyp)
        elif estimator.dif_type == 2:
            unit, name = _get_detector_unit(SHDetType.let_bdo2016, estimator.geotyp)
        elif estimator.dif_type == 3:
            unit, name = _get_detector_unit(SHDetType.angle_bdo2016, estimator.geotyp)
        else:
            unit, name = _get_detector_unit(estimator.dif_type, estimator.geotyp)

    return unit, name


def _bintyp(n):
    """
    Calculates type of binning based on number of bins.

    We follow the convention that positive number of bins means linear binning,
    while the negative - logarithmic.

    :param n: number of bins
    :return: MeshAxis.BinningType.linear or MeshAxis.BinningType.logarithmic
    """
    return MeshAxis.BinningType.linear if n > 0 else MeshAxis.BinningType.logarithmic


def _get_detector_unit(detector_type, geotyp):
    """
    TODO
    :param detector_type:
    :param geotyp:
    :return:
    """
    if geotyp == SHGeoType.zone:
        dose_units = ("MeV/primary", "Dose*volume")
        dose_gy_units = ("J", "Dose*volume")
        alanine_units = ("MeV/primary", "Alanine RE*Dose*volume")
        alanine_gy_units = ("J", "Alanine RE*Dose*volume")
    else:
        dose_units = (" MeV/g/primary", "Dose")
        dose_gy_units = ("Gy", "Dose")
        alanine_units = ("MeV/g/primary", "Alanine RE*Dose")
        alanine_gy_units = ("Gy", "Alanine RE*Dose")

    # TODO add more units, move to shieldhit/detector package
    _detector_units = {
        SHDetType.none: ("(nil)", "None"),
        SHDetType.energy: ("MeV/primary", "Energy"),
        SHDetType.fluence: ("cm^-2/primary", "Fluence"),
        SHDetType.crossflu: ("cm^-2/primary", "Planar fluence"),
        SHDetType.letflu: ("MeV/cm", "LET fluence"),
        SHDetType.dose: dose_units,
        SHDetType.dose_gy_bdo2016: dose_gy_units,
        SHDetType.dlet: ("keV/um", "dose-averaged LET"),
        SHDetType.tlet: ("keV/um", "track-averaged LET"),
        SHDetType.avg_energy: ("MeV/nucleon", "Average kinetic energy"),
        SHDetType.avg_beta: ("(dimensionless)", "Average beta"),
        SHDetType.material: ("(nil)", "Material number"),
        SHDetType.alanine: alanine_units,
        SHDetType.alanine_gy_bdo2016: alanine_gy_units,
        SHDetType.counter: ("/primary", "Particle counter"),
        SHDetType.pet: ("/primary", "PET isotopes"),
        SHDetType.dletg: ("keV/um", "dose-averaged LET"),
        SHDetType.tletg: ("keV/um", "track-averaged LET"),
        SHDetType.q: ("(nil)", "beam quality Q"),
        SHDetType.flu_char: ("cm^-2/primary", "Charged particle fluence"),
        SHDetType.flu_neut: ("cm^-2/primary", "Neutral particle fluence"),
        SHDetType.flu_neqv: ("cm^-2/primary", "1 MeV eqv. neutron fluence"),
        SHDetType.let_bdo2016: ("keV/um", "LET"),
        SHDetType.angle_bdo2016: ("radians", "Angle"),
        SHDetType.zone: ("(dimensionless)", "Zone#"),
        SHDetType.medium: ("(dimensionless)", "Medium#"),
        SHDetType.rho: ("g/cm^3", "Density"),
        SHDetType.kinetic_energy: ("MeV", "Kinetic energy"),
    }
    return _detector_units.get(detector_type, ("(nil)", "(nil)"))


def read_next_token(f):
    """
    returns a tuple with 4 elements:
    0: payload id
    1: payload dtype string
    2: payload number of elements
    3: payload itself
    f is an open and readable file pointer.
    returns None if no token was found / EOF
    """
    tag = np.dtype([('pl_id', '<u8'),
                    ('pl_type', 'S8'),
                    ('pl_len', '<u8')])

    x1 = np.fromfile(f, dtype=tag, count=1)  # read the data into numpy

    if x1.size == 0:
        return None
    else:
        pl_id = x1['pl_id'][0]
        pl_type = x1['pl_type'][0]
        pl_len = x1['pl_len'][0]
        try:
            pl = np.fromfile(f,
                             dtype=pl_type,
                             count=pl_len)  # read the data into numpy
            return pl_id, pl_type, pl_len, pl
        except TypeError:
            return None


def _postprocess(estimator, nscale):
    """normalize result if we need that."""
    for page in estimator.pages:
        if page.dettyp not in (SHDetType.dlet, SHDetType.tlet, SHDetType.letflu, SHDetType.dletg, SHDetType.tletg,
                               SHDetType.avg_energy, SHDetType.avg_beta, SHDetType.material, SHDetType.q):
            if estimator.number_of_primaries != 0:  # geotyp = GEOMAP will have 0 projectiles simulated
                page.data_raw /= np.float64(estimator.number_of_primaries)
                page.error_raw /= np.float64(estimator.number_of_primaries)

    if nscale != 1:
        # scale with number of particles given by user
        for page in estimator.pages:
            page.data_raw *= np.float64(nscale)
            page.error_raw *= np.float64(nscale)

        # rescaling with particle number means also unit change for some estimators
        # from per particle to Grey - this is why we override detector type
        for page in estimator.pages:
            page.data_raw *= np.float64(nscale)
            page.error_raw *= np.float64(nscale)

            if page.dettyp == SHDetType.dose:
                page.dettyp = SHDetType.dose_gy_bdo2016
            if page.dettyp == SHDetType.alanine:
                page.dettyp = SHDetType.alanine_gy_bdo2016
            # for the same reason as above we change units
            if page.dettyp in (SHDetType.dose_gy_bdo2016, SHDetType.alanine_gy_bdo2016):
                # 1 megaelectron volt / gram = 1.60217662 x 10-10 Gy
                MeV_g = np.float64(1.60217662e-10)
                page.data_raw *= MeV_g
                page.error_raw *= MeV_g
                page.unit, page.name = _get_detector_unit(page.dettyp, estimator.geotyp)



from enum import IntEnum

from pymchelper.writers.excel import ExcelWriter
from pymchelper.writers.inspector import Inspector
from pymchelper.writers.plots import PlotDataWriter, ImageWriter
from pymchelper.writers.shieldhit import TxtWriter
from pymchelper.writers.sparse import SparseWriter
from pymchelper.writers.trip98cube import TRiP98CubeWriter
from pymchelper.writers.trip98ddd import TRiP98DDDWriter
from pymchelper.writers.hdf import HdfWriter
from pymchelper.writers.json import JsonWriter
from pymchelper.writers.mcpl import MCPLWriter


class Converters(IntEnum):
    """
    Available converters
    """
    txt = 0
    plotdata = 1
    image = 3
    tripcube = 4
    tripddd = 5
    excel = 6
    sparse = 7
    inspect = 8
    hdf = 9
    json = 10
    mcpl = 11

    @classmethod
    def _converter_mapping(cls, item):
        return {
            cls.txt: TxtWriter,
            cls.plotdata: PlotDataWriter,
            cls.image: ImageWriter,
            cls.tripcube: TRiP98CubeWriter,
            cls.tripddd: TRiP98DDDWriter,
            cls.excel: ExcelWriter,
            cls.sparse: SparseWriter,
            cls.inspect: Inspector,
            cls.hdf: HdfWriter,
            cls.json: JsonWriter,
            cls.mcpl: MCPLWriter,
        }.get(item)

    @classmethod
    def fromname(cls, name):
        return cls._converter_mapping(Converters[name])

    @classmethod
    def fromnumber(cls, number):
        return cls._converter_mapping[Converters(number)]



import copy
from enum import IntEnum
import logging
from typing import Optional

import numpy as np
from pymchelper.axis import MeshAxis, AxisId

logger = logging.getLogger(__name__)


class ErrorEstimate(IntEnum):
    """
    When averaging data multiple files we could estimate statistical error of scored quantity.
    Such error can be calculated as: none (error information missing), standard error or standard deviation.
    """
    none = 0
    stderr = 1
    stddev = 2


class Estimator:
    """
    Estimator data including scoring mesh description.

    This class handles in universal way data generated with MC code.
    It includes data (``data`` and ``data_raw`` fields) and optional errors (``error`` and ``error_raw``).
    Estimator holds also up to 3 binning axis (``x``, ``y`` and ``z`` fields).
    Scored quantity can be assigned a ``name`` (i.e. dose) and ``unit`` (i.e. Gy).
    Several other fields are also used:
    - nstat: number of simulated histories
    - counter: number of files read to construct detector object
    - corename: common core part of input files defining a name of detector
    - error_type: none, stderr or stddev - error type

    Estimator data can be either read from the file (see ``fromfile`` method in ``input_output`` module
    or constructed directly:

    >>> d = Estimator()
    >>> d.x = MeshAxis(n=2, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> d.x.data
    array([2.5, 7.5])
    >>> d.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> d.y.data
    array([  25.,   75.,  125.])
    >>> d.z = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> d.z.data
    array([0.5])
    """

    def __init__(self):
        """
        Create dummy estimator object.
        >>> e = Estimator()
        """
        self.x = MeshAxis(n=1,
                          min_val=float("NaN"),
                          max_val=float("NaN"),
                          name="",
                          unit="",
                          binning=MeshAxis.BinningType.linear)
        self.y = self.x
        self.z = self.x

        self.number_of_primaries = 0  # number of histories simulated
        self.file_counter = 0  # number of files read
        self.file_corename = ""  # common core for paths of contributing files
        self.file_format = ""  # binary file format of the input files
        self.error_type = ErrorEstimate.none
        self.geotyp = None  # MSH, CYL, etc...

        self.pages = ()  # empty tuple of pages at the beginning

    def add_page(self, page):
        """
        Add a page to the estimator object.
        New copy of page is made and page estimator pointer is set to the estimator object holding this page.
        :param page:
        :return: None
        """
        new_page = copy.deepcopy(page)
        new_page.estimator = self
        self.pages += (new_page,)

    def axis(self, axis_id : int) -> Optional[MeshAxis]:
        """
        Mesh axis selector method based on integer id's.

        Instead of getting mesh axis data by calling `d.x`, `d.y` or `d.z` (assuming `d` an object of `Detector`
        class) we can get that data by calling `d.axis(0)`, `d.axis(1)` or `d.axis(2)`. See for example:
        >>> d = Estimator()
        >>> d.x = MeshAxis(n=2, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.z = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> d.axis(1)
        MeshAxis(n=3, min_val=0.0, max_val=150.0, name='Y', unit='cm', binning=<BinningType.linear: 0>)

        :param axis_id: axis id (0, 1 or 2)
        :return: MeshAxis object
        """
        if axis_id == AxisId.x:
            return self.x
        elif axis_id == AxisId.y:
            return self.y
        elif axis_id == AxisId.z:
            return self.z
        return None

    @property
    def dimension(self) -> int:
        """
        Let's take again detector d with YZ scoring.
        >>> e = Estimator()
        >>> e.x = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.z = MeshAxis(n=2, min_val=0.0, max_val=2.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.dimension
        2

        :return: number of axes (among X,Y,Z) which have more than one bin
        """
        return 3 - (self.x.n, self.y.n, self.z.n).count(1)


def average_with_nan(estimator_list, error_estimate=ErrorEstimate.stderr):
    """
    Calculate average estimator object, excluding malformed data (NaN) from averaging.
    :param estimator_list:
    :param error_estimate:
    :return:
    """
    # TODO add compatibility check
    if not estimator_list:
        return None
    result = copy.deepcopy(estimator_list[0])
    for page_no, page in enumerate(result.pages):
        page.data_raw = np.nanmean([estimator.pages[page_no].data_raw for estimator in estimator_list], axis=0)
    result.file_counter = len(estimator_list)
    if result.file_counter > 1 and error_estimate != ErrorEstimate.none:
        # s = stddev = sqrt(1/(n-1)sum(x-<x>)**2)
        # s : corrected sample standard deviation
        for page_no, page in enumerate(result.pages):
            page.error_raw = np.nanstd([estimator.pages[page_no].data_raw for estimator in estimator_list],
                                       axis=0, ddof=1)

        # if user requested standard error then we calculate it as:
        # S = stderr = stddev / sqrt(n), or in other words,
        # S = s/sqrt(N) where S is the corrected standard deviation of the mean.
        if error_estimate == ErrorEstimate.stderr:
            for page in result.pages:
                page.error_raw /= np.sqrt(result.file_counter)  # np.sqrt() always returns np.float64
    else:
        for page in result.pages:
            page.error_raw = np.zeros_like(page.data_raw)
    return result



# $Id: bmath.py 3739 2016-03-22 08:14:24Z bnv $
#
# Copyright and User License
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Copyright Vasilis.Vlachoudis@cern.ch for the
# European Organization for Nuclear Research (CERN)
#
# All rights not expressly granted under this license are reserved.
#
# Installation, use, reproduction, display of the
# software ("flair"), in source and binary forms, are
# permitted free of charge on a non-exclusive basis for
# internal scientific, non-commercial and non-weapon-related
# use by non-profit organizations only.
#
# For commercial use of the software, please contact the main
# author Vasilis.Vlachoudis@cern.ch for further information.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#
# DISCLAIMER
# ~~~~~~~~~~
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY, OF
# SATISFACTORY QUALITY, AND FITNESS FOR A PARTICULAR PURPOSE
# OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDERS AND THE
# AUTHORS MAKE NO REPRESENTATION THAT THE SOFTWARE AND
# MODIFICATIONS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# LIMITATION OF LIABILITY
# ~~~~~~~~~~~~~~~~~~~~~~~
# THE COPYRIGHT HOLDERS AND THE AUTHORS SHALL HAVE NO
# LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
# CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES OF ANY
# CHARACTER INCLUDING, WITHOUT LIMITATION, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR PROFITS,
# OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY
# OF CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT
# LIABILITY OR OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH

# Author:	Vasilis.Vlachoudis@cern.ch
# Date:	15-May-2004
from __future__ import generators

import math
import random

import pymchelper.flair.common.rexx as rexx

__author__ = "Vasilis Vlachoudis"
__email__ = "Vasilis.Vlachoudis@cern.ch"

# Accuracy for comparison operators
_accuracy = 1E-15

# Formating
_format = "%12g"


# -------------------------------------------------------------------------------
def Cmp0(x):
    return abs(x) < _accuracy


# -------------------------------------------------------------------------------
# range of floating point numbers
# -------------------------------------------------------------------------------
def frange(start, stop, step):
    """range(start,stop,step) for floating point numbers"""
    x = start
    if step < 0.0:
        while x > stop:
            yield x
            x += step
    else:
        while x < stop:
            yield x
            x += step


# -------------------------------------------------------------------------------
# limit in number within a range
# -------------------------------------------------------------------------------
def limit(min_, num, max_):
    '''
    limit a number within a specific range
    :param min_:
    :param num:
    :param max_:
    :return:
    '''
    return max(min(num, max_), min_)


# -------------------------------------------------------------------------------
# dms - degrees from degrees, minutes, seconds
# -------------------------------------------------------------------------------
def dms(d, m, s):
    return d + m / 60.0 + s / 3600.0


# -------------------------------------------------------------------------------
# d2s - degrees to string
# D2S(angle[,"H"|"M"|"D"|"N"])
# -------------------------------------------------------------------------------
def d2s(ang, fmt=""):
    fmt.capitalize()
    if ang < 0.0:
        neg = "-"
        ang = -ang
    else:
        neg = ""

    ang = round(ang * 360000) / 100
    SS = "%05.2f" % (math.fmod(ang, 60))
    ang = int(ang / 60.0)
    MM = "%02d" % (ang % 60)
    HH = neg + str(ang / 60)

    if fmt == "H":
        return HH + "h" + MM + "m" + SS + "s"
    if fmt == "M":
        return HH + "h" + MM + "m"
    if fmt == "D":
        return HH + " " + MM + "'" + SS + '"'
    if fmt == "N":
        return HH + ":" + MM
    return HH + ":" + MM + ":" + SS


# -------------------------------------------------------------------------------
# Format a number to fit in the minimum space
# -------------------------------------------------------------------------------
def format_number(number, length=10, useExp=False, useD=False):
    """
    Format a number to fit in the minimum space given by length
    """

    _MAXLEN = 22

    # Convert number to string
    # XXX str cuts at 12 digits, repr shows everything but
    # numbers like e.g 9.2 will be converted to 9.1999999999999
    # What should I do
    # Fields also in the CardWidget are converted with str and
    # are cut at 12 digits!!!!
    if isinstance(number, float) or isinstance(number, int):
        number = repr(number).upper()
    else:
        number = str(number).strip().upper()
        if not rexx.datatype(number, "N"):
            return number

    if useD:
        number = number.replace("E", "D")
        expE = "D"
    else:
        number = number.replace("D", "E")
        expE = "E"

    if len(number) < length:
        hasExp = (number.find(expE) >= 0)
        if useExp:
            if hasExp:
                return number
        elif number.find(".") >= 0 or hasExp:
            return number

    if number == "0":
        if useExp:
            return "0.%s0" % (expE)
        else:
            return "0.0"

    if length < 5 or length > _MAXLEN:
        raise Exception("Format invalid length")

    # Dissect the number. It is in the normal Rexx format.
    try:
        (mantissa, exponent) = number.split(expE)
        if exponent == '':
            exponent = 0
        else:
            exponent = int(exponent)
    except Exception:
        mantissa = number
        exponent = 0

    if mantissa[0] == '-':
        sgn = True
        mantissa = mantissa[1:]
    elif mantissa[0] == '+':
        sgn = False
        mantissa = mantissa[1:]
    else:
        sgn = False

    try:
        (befo, afte) = mantissa.split(".")
    except Exception:
        befo = mantissa
        afte = ""

    # Count from the left for the decimal point.
    point = len(befo)

    # Make this a number without a point.
    integer = befo + afte

    # Remove leading zeros
    for p in range(len(integer)):
        if integer[p] != '0':
            if p > 0:
                integer = integer[p:]
                point -= p
            break
    else:
        if useExp:
            return "0.%s0" % (expE)
        else:
            return "0.0"

    # ... and trailing
    for p in range(len(integer) - 1, 0, -1):
        if integer[p] != '0':
            integer = integer[0:p + 1]
            break

    exponent += point

    # Cannot handle more than _MAXLEN digits
    lint = len(integer)
    if lint > _MAXLEN:
        r = integer[_MAXLEN]
        integer = integer[0:_MAXLEN]
        if r >= '5':
            integer = str(int(integer) + 1)
            if len(integer) > lint:
                exponent += 1
                if len(integer) > _MAXLEN:
                    integer = integer[0:_MAXLEN]

    # Now the number is described by:
    #   sgn 0.integer "E" exponent

    # Make space for sign
    if sgn:
        length -= 1

    while True:
        # Minimum length representation of a number
        # Length = Length of integer
        #       + 1 for Dot if needed (no exponent)
        #       + (2-4) for exponent
        #  exponent can be in the following forms
        #       nothing if dot can placed inside integer
        #       E# 2
        #       E## 3
        #       E-# 3
        #       E-## 4
        #   integer is given as  0.integer
        # say("\ninteger=",integer,len(integer),length)
        # say("exponent=",exponent)
        lint = len(integer)
        if useExp:
            mNum = "%s%s%d" % (rexx.insert(".", integer, 1), expE, exponent - 1)
        elif exponent == -2:
            mNum = ".00%s" % (integer)
        elif exponent == -1:
            mNum = ".0%s" % (integer)
        elif exponent == 0:
            mNum = ".%s" % (integer)
        elif exponent == 1:
            mNum = rexx.insert(".", integer, 1)
        elif exponent == length:
            mNum = "%s%s" % (integer, "0" * (length - lint))
        elif exponent > 1 and exponent <= lint:
            mNum = rexx.insert(".", integer, exponent)
        elif exponent > 1 and exponent <= lint + 2:
            if exponent > lint:
                mNum = "%s%s." % (integer, "0" * (exponent - lint))
            else:
                mNum = "%s." % (integer.ljust(exponent))
        elif exponent > lint and exponent + 1 < length:
            mNum = "%s%s." % (integer, "0" * (exponent - lint))
        else:
            mNum = "%s%s%d" % (rexx.insert(".", integer, 1), expE, exponent - 1)

        diff = len(mNum) - length
        if diff <= 0:
            break
        elif diff <= 2:
            r = integer[-1]
            integer = integer[0:-1]
        else:
            r = integer[-diff]
            integer = integer[0:-diff]

        if r >= '5':
            lint = len(integer)
            if lint == 0:
                integer = 0
            integer = str(int(integer) + 1)
            if len(integer) > lint:
                exponent += 1

        # Remove trailing zeros
        for p in range(len(integer) - 1, -1, -1):
            if integer[p] != '0':
                integer = integer[0:p + 1]
                break
        else:
            if useExp:
                return "0.%s0" % (expE)
            else:
                return "0.0"

    if sgn:
        mNum = "-%s" % (mNum)
    return mNum


# ==============================================================================
# Dangerous dictionary that unknown keys return a user default value
# Use it with care
# ==============================================================================
class DefaultDict(dict):
    def __init__(self, default=None):
        self._default = default

    # ----------------------------------------------------------------------
    def __getitem__(self, key):
        return self.get(key, self._default)


# ==============================================================================
# Unknown keys will return 0.0
# ==============================================================================
class ZeroDict(DefaultDict):
    def __init__(self):
        DefaultDict.__init__(self, 0.0)


# ==============================================================================
# Unknown keys will return 0.0
# ==============================================================================
class ZeroIntDict(DefaultDict):
    def __init__(self):
        DefaultDict.__init__(self, 0)


# ===============================================================================
# Vector class
# Inherits from List
# ===============================================================================
class Vector(list):
    """Vector class"""

    # ----------------------------------------------------------------------
    def __init__(self, x=3, *args):
        """Create a new vector,
        Vector(size), Vector(list), Vector(x,y,z,...)"""

        if isinstance(x, int) and not args:
            for i in range(x):
                self.append(0.0)
        elif isinstance(x, list) or isinstance(x, tuple):
            for i in x:
                self.append(float(i))
        else:
            self.append(float(x))
            for i in args:
                self.append(float(i))

    # ----------------------------------------------------------------------
    def set(self, x, y, z=None):
        """Set vector"""
        self[0] = x
        self[1] = y
        if z:
            self[2] = z

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "[%s]" % (", ".join([(_format % (x)).strip() for x in self]))

    # ----------------------------------------------------------------------
    def eq(self, v, acc=_accuracy):
        """Test for equality with vector v within accuracy"""
        if len(self) != len(v):
            return False
        s2 = 0.0
        for i in range(len(self)):
            s2 += (self[i] - v[i])**2
        return s2 <= acc**2

    def __eq__(self, v):
        return self.eq(v)

    # ----------------------------------------------------------------------
    def __neg__(self):
        """Negate vector"""
        new = Vector(len(self))
        for i in range(len(self)):
            new[i] = -self[i]
        return new

    # ----------------------------------------------------------------------
    def __add__(self, v):
        """Add 2 vectors"""
        size = min(len(self), len(v))
        new = Vector(size)
        for i in range(size):
            new[i] = self[i] + v[i]
        return new

    # ----------------------------------------------------------------------
    def __iadd__(self, v):
        """Add vector v to self"""
        for i in range(min(len(self), len(v))):
            self[i] += v[i]
        return self

    # ----------------------------------------------------------------------
    def __sub__(self, v):
        """Subtract 2 vectors"""
        size = min(len(self), len(v))
        new = Vector(size)
        for i in range(size):
            new[i] = self[i] - v[i]
        return new

    # ----------------------------------------------------------------------
    def __isub__(self, v):
        """Subtract vector v from self"""
        for i in range(min(len(self), len(v))):
            self[i] -= v[i]
        return self

    # ----------------------------------------------------------------------
    # Scale or Dot product
    # ----------------------------------------------------------------------
    def __mul__(self, v):
        """scale*Vector() or Vector()*Vector() - Scale vector or dot product"""
        if isinstance(v, list):
            return self.dot(v)
        else:
            return Vector([x * v for x in self])

    # ----------------------------------------------------------------------
    # Scale or Dot product
    # ----------------------------------------------------------------------
    def __rmul__(self, v):
        """scale*Vector() or Vector()*Vector() - Scale vector or dot product"""
        if isinstance(v, Vector):
            return self.dot(v)
        else:
            return Vector([x * v for x in self])

    # ----------------------------------------------------------------------
    # Divide by floating point
    # ----------------------------------------------------------------------
    def __div__(self, b):
        return Vector([x / b for x in self])

    # ----------------------------------------------------------------------
    def __xor__(self, v):
        """Cross product"""
        return self.cross(v)

    # ----------------------------------------------------------------------
    def dot(self, v):
        """Dot product of 2 vectors"""
        size = min(len(self), len(v))
        s = 0.0
        for i in range(size):
            s += self[i] * v[i]
        return s

    # ----------------------------------------------------------------------
    def cross(self, v):
        """Cross product of 2 vectors"""
        if len(self) == 3:
            return Vector(self[1] * v[2] - self[2] * v[1], self[2] * v[0] - self[0] * v[2],
                          self[0] * v[1] - self[1] * v[0])
        elif len(self) == 2:
            return self[0] * v[1] - self[1] * v[0]
        else:
            raise Exception("Cross product needs 2d or 3d vectors")

    # ----------------------------------------------------------------------
    def length2(self):
        """Return length squared of vector"""
        s = 0.0
        for i in range(len(self)):
            s += self[i]**2
        return s

    # ----------------------------------------------------------------------
    def length(self):
        """Return length of vector"""
        s = 0.0
        for i in range(len(self)):
            s += self[i]**2
        return math.sqrt(s)

    __abs__ = length

    # ----------------------------------------------------------------------
    def arg(self):
        """return vector angle"""
        return math.atan2(self[1], self[0])

    # ----------------------------------------------------------------------
    def norm(self):
        """Normalize vector and return length"""
        s_length = self.length()
        if s_length > 0.0:
            invlen = 1.0 / s_length
            for i in range(len(self)):
                self[i] *= invlen
        return s_length

    normalize = norm

    # ----------------------------------------------------------------------
    def unit(self):
        """return a unit vector"""
        v = self.clone()
        v.norm()
        return v

    # ----------------------------------------------------------------------
    def clone(self):
        """Clone vector"""
        return Vector(self)

    # ----------------------------------------------------------------------
    def x(self):
        return self[0]

    def y(self):
        return self[1]

    def z(self):
        return self[2]

    # ----------------------------------------------------------------------
    def orthogonal(self):
        """return a vector orthogonal to self"""
        xx = abs(self.x())
        yy = abs(self.y())

        if len(self) >= 3:
            zz = abs(self.z())
            if xx < yy:
                if xx < zz:
                    return Vector(0.0, self.z(), -self.y())
                else:
                    return Vector(self.y(), -self.x(), 0.0)
            else:
                if yy < zz:
                    return Vector(-self.z(), 0.0, self.x())
                else:
                    return Vector(self.y(), -self.x(), 0.0)
        else:
            return Vector(-self.y(), self.x())

    # ----------------------------------------------------------------------
    def direction(self, zero=_accuracy):
        """return containing the direction
        if normalized with any of the axis"""

        v = self.clone()
        length = v.norm()
        if abs(length) <= zero:
            return "O"
        if abs(v[0] - 1.0) < zero:
            return "X"
        elif abs(v[0] + 1.0) < zero:
            return "-X"
        elif abs(v[1] - 1.0) < zero:
            return "Y"
        elif abs(v[1] + 1.0) < zero:
            return "-Y"
        elif abs(v[2] - 1.0) < zero:
            return "Z"
        elif abs(v[2] + 1.0) < zero:
            return "-Z"
        else:
            # nothing special about the direction, return N
            return "N"

    # ----------------------------------------------------------------------
    # Set the vector directly in polar coordinates
    # @param ma magnitude of vector
    # @param ph azimuthal angle in radians
    # @param th polar angle in radians
    # ----------------------------------------------------------------------
    def setPolar(self, ma, ph, th):
        sf = math.sin(ph)
        cf = math.cos(ph)
        st = math.sin(th)
        ct = math.cos(th)
        self[0] = ma * st * cf
        self[1] = ma * st * sf
        self[2] = ma * ct

    # ----------------------------------------------------------------------
    # @return the azimuth angle.
    # ----------------------------------------------------------------------
    def phi(self):
        if Cmp0(self.x()) and Cmp0(self.y()):
            return 0.0
        return math.atan2(self.y(), self.x())

    # ----------------------------------------------------------------------
    # @return the polar angle.
    # ----------------------------------------------------------------------
    def theta(self):
        if Cmp0(self.x()) and Cmp0(self.y()) and Cmp0(self.z()):
            return 0.0
        return math.atan2(self.perp(), self.z())

    # ----------------------------------------------------------------------
    # @return cosine of the polar angle.
    # ----------------------------------------------------------------------
    def cosTheta(self):
        ptot = self.length()
        if Cmp0(ptot):
            return 1.0
        else:
            return self.z() / ptot

    # ----------------------------------------------------------------------
    # @return the transverse component squared
    #   (R^2 in cylindrical coordinate system).
    # ----------------------------------------------------------------------
    def perp2(self):
        return self.x() * self.x() + self.y() * self.y()

    # ----------------------------------------------------------------------
    # @return the transverse component
    # (R in cylindrical coordinate system).
    # ----------------------------------------------------------------------
    def perp(self):
        return math.sqrt(self.perp2())


# -------------------------------------------------------------------------------
# Basic 3D Vectors
# -------------------------------------------------------------------------------
Vector.O = Vector(0.0, 0.0, 0.0)  # noqa: E741
Vector.X = Vector(1.0, 0.0, 0.0)
Vector.Y = Vector(0.0, 1.0, 0.0)
Vector.Z = Vector(0.0, 0.0, 1.0)


# -------------------------------------------------------------------------------
# Return a random 3D vector
# -------------------------------------------------------------------------------
def random3D():
    cosTheta = 2.0 * random.random() - 1.0
    sinTheta = math.sqrt(1.0 - cosTheta**2)
    phi = 2.0 * math.pi * random.random()
    return Vector(math.cos(phi) * sinTheta, math.sin(phi) * sinTheta, cosTheta)


# ------------------------------------------------------------------------------
# Return a random nolor
# ------------------------------------------------------------------------------
def rndColor(x):
    def rnd(zw):
        w = zw & 0xffff
        z = (zw >> 16) & 0xffff
        z = 36969 * (z & 0xffff) + (z >> 16)
        w = 18000 * (w & 0xffff) + (w >> 16)
        return (z << 16) + (w & 0xffff)

    x = rnd(x)
    R = (x % 224) + 16
    x = rnd(x)
    G = (x % 224) + 16
    x = rnd(x)
    B = (x % 224) + 16
    return R << 16 | G << 8 | B


# ===============================================================================
# Matrix class
# Use 4x4 matrix for vector transformations
# ===============================================================================
class Matrix(list):
    """Matrix 4x4 used for vector transformations"""

    # ----------------------------------------------------------------------
    def __init__(self, rows=4, cols=-1, type=0):
        """
        Matrix(rows=4, cols=-1, type=0|1)
        if rows is integer then
            Create a matrix rows x cols either
            zero(type=0) or unary(type=1)
        elif rows is a list of lists
            create a matrix from a double-list
        """
        if isinstance(rows, list):
            lst = rows
            self.rows = len(lst)
            self.extend([[]] * self.rows)
            if isinstance(lst[0], list):
                self.cols = len(lst[0])
                for i in range(self.rows):
                    self[i] = lst[i][:]
                    if len(self[i]) != self.cols:
                        raise Exception("Not a valid double-list for a matrix")
            else:
                self.cols = 1
                for i in range(self.rows):
                    self[i] = [lst[i]]
        else:
            if rows < 2:
                raise Exception("Array size too small")
            if cols < 0:
                cols = rows
            self.rows = rows
            self.cols = cols
            self += [[]] * rows
            if type == 1:
                self.unary()
            else:
                self.zero()

    # ----------------------------------------------------------------------
    # Create a diagonal square matrix from a list
    # ----------------------------------------------------------------------
    @staticmethod
    def diagonal(lst):
        m = Matrix(len(lst), type=0)
        i = 0
        for item in lst:
            m[i][i] = item
            i += 1
        return m

    # ----------------------------------------------------------------------
    # append row
    # ----------------------------------------------------------------------
    def append(self, col):
        list.append(self, col)
        self.rows += 1

    # ----------------------------------------------------------------------
    @staticmethod
    def translate(x, y=0.0, z=0.0):
        """m = Matrix.translate(x,y,z|vector)
        @return a translation matrix"""
        m = Matrix(4, type=1)
        if isinstance(x, list) or isinstance(x, tuple):
            m[0][3] = x[0]
            m[1][3] = x[1]
            m[2][3] = x[2]
        else:
            m[0][3] = x
            m[1][3] = y
            m[2][3] = z
        return m

    # ----------------------------------------------------------------------
    @staticmethod
    def scale(sx, sy=None, sz=None):
        """m = Matrix.scale(scale|vector)
           @return a scaling matrix"""
        m = Matrix(4, type=1)
        if sy is None:
            sy = sx
        if sz is None:
            sz = sx
        if isinstance(sx, list) or isinstance(sx, tuple):
            m[0][0] = sx[0]
            m[1][1] = sx[1]
            m[2][2] = sx[2]
        else:
            m[0][0] = sx
            m[1][1] = sy
            m[2][2] = sz
        return m

    # ----------------------------------------------------------------------
    def zero(self):
        """Zero matrix"""
        for i in range(self.rows):
            self[i] = [0.0] * self.cols

    # ----------------------------------------------------------------------
    def unary(self):
        """Unary matrix"""
        self.zero()
        for i in range(min(self.rows, self.cols)):
            self[i][i] = 1.0

    # ----------------------------------------------------------------------
    # Create a transformation matrix from 3 normalized vectors
    # and optionally a translation
    # ----------------------------------------------------------------------
    def make(self, X, Y, Z=None, T=None):
        """Create a transformation matrix from 3 normalized vectors"""
        self.unary()
        if self.rows in (3, 4) and self.cols == self.rows:
            if Z is None:
                Z = X ^ Y
                Z.normalize()
            for i in range(3):
                self[0][i] = X[i]
                self[1][i] = Y[i]
                self[2][i] = Z[i]
                if T is not None and self.rows == 4:
                    self[i][3] = T[i]
        else:
            raise Exception("Matrix.make() works only " "on Matrix(3x3) or Matrix(4x4)")

    # ----------------------------------------------------------------------
    def __str__(self):
        """Multiline string representation of matrix"""
        s = ""
        for i in range(self.rows):
            if i == 0:
                first = "/"
                last = "\\"
            elif i == self.rows - 1:
                first = "\\"
                last = "/"
            else:
                first = last = "|"
            s += first
            for j in range(self.cols):
                s += " " + _format % self[i][j]
            s += " " + last + "\n"
        return s

    # ----------------------------------------------------------------------
    def writeOctave(self, filename, name):
        """Write an octave matrix file"""
        f = open(filename, "w")
        f.write("# bmath.Matrix\n")
        f.write("# name: %s\n" % (name))
        f.write("# type: matrix\n")
        f.write("# rows: %d\n" % (self.rows))
        f.write("# columns: %d\n" % (self.cols))
        for i in range(self.rows):
            for j in range(self.cols):
                f.write("%s " % (str(self[i][j])))
            f.write("\n")
        f.close()

    # ----------------------------------------------------------------------
    def T(self):
        """@return transpose matrix"""
        m = Matrix(self.cols, self.rows)
        for i in range(self.rows):
            for j in range(self.cols):
                m[j][i] = self[i][j]
        return m

    transpose = T

    # ----------------------------------------------------------------------
    def trace(self):
        """Return trace of matrix (sum of diagonal elements)"""
        t = 0.0
        for i in range(min(self.rows, self.cols)):
            t += self[i][i]
        return t

    # ----------------------------------------------------------------------
    def __eq__(self, m):
        """Test for equality of 2 matrices"""
        if self.rows != m.rows or self.cols != m.cols:
            return False
        for i in range(self.rows):
            for j in range(self.cols):
                if abs(self[i][j] - m[i][j]):
                    return False
        return True

    # ----------------------------------------------------------------------
    # Create a rotation matrix around one axis
    #   X = 0
    #   Y = 1
    #   Z = 2
    # or an arbitrary vector
    # ----------------------------------------------------------------------
    def rotate(self, angle, axis):
        """Add rotation elements to the matrix around one axis
        Axis X=0, Y=1, Z=2, or an arbitrary one given by vector axis"""
        self.unary()

        c = math.cos(angle)
        s = math.sin(angle)

        if isinstance(axis, int):
            m1 = ((axis + 1) % 3) + 1
            m2 = m1 % 3
            m1 -= 1

            self[m1][m1] = c
            self[m2][m2] = c
            self[m1][m2] = -s
            self[m2][m1] = s

        elif isinstance(axis, Vector):
            length = axis.length()
            x = axis[0] / length
            y = axis[1] / length
            z = axis[2] / length

            c1 = 1 - c
            self[0][0] = x * x + (1 - x * x) * c
            self[0][1] = x * y * c1 - z * s
            self[0][2] = x * z * c1 + y * s

            self[1][0] = x * y * c1 + z * s
            self[1][1] = y * y + (1 - y * y) * c
            self[1][2] = y * z * c1 - x * s

            self[2][0] = x * z * c1 - y * s
            self[2][1] = y * z * c1 + x * s
            self[2][2] = z * z + (1 - z * z) * c

    # ----------------------------------------------------------------------
    @staticmethod
    def rotX(angle):
        """m = Matrix.rotX(angle) - Return a rotation matrix around X"""
        m = Matrix(4, type=1)
        m.rotate(angle, 0)
        return m

    # ----------------------------------------------------------------------
    @staticmethod
    def rotY(angle):
        """m = Matrix.rotY(angle) - Return a rotation matrix arround Y"""
        m = Matrix(4, type=1)
        m.rotate(angle, 1)
        return m

    # ----------------------------------------------------------------------
    @staticmethod
    def rotZ(angle):
        """m = Matrix.rotZ(angle) - Return a rotation matrix arround Z"""
        m = Matrix(4, type=1)
        m.rotate(angle, 2)
        return m

    # ----------------------------------------------------------------------
    def getEulerRotation(self):
        # ROTX(x) * ROTY(y) * ROTZ(z)
        #  cos(z)*cos(y)
        #  sin(z)*cos(y)
        #        -sin(y)
        # -sin(z)*cos(x)+cos(z)*sin(y)*sin(x)
        #  cos(z)*cos(x)+sin(z)*sin(y)*sin(x)
        #                       cos(y)*sin(x)
        #  sin(z)*sin(x)+cos(z)*sin(y)*cos(x)
        # -cos(z)*sin(x)+sin(z)*sin(y)*cos(x)
        #                       cos(y)*cos(x)
        rx = math.atan2(self[1][2], self[2][2])
        ry = -math.asin(self[0][2])
        rz = math.atan2(self[0][1], self[0][0])
        return rx, ry, rz

    # ----------------------------------------------------------------------
    @staticmethod
    def eulerRotation(rx, ry, rz):
        # ROTX(x) * ROTY(y) * ROTZ(z)
        m = Matrix(4, type=1)
        cx = math.cos(rx)
        cy = math.cos(ry)
        cz = math.cos(rz)
        sx = math.sin(rx)
        sy = math.sin(ry)
        sz = math.sin(rz)

        row = m[0]
        row[0] = cz * cy
        row[1] = sz * cy
        row[2] = -sy

        row = m[1]
        row[0] = -sz * cx + cz * sy * sx
        row[1] = cz * cx + sz * sy * sx
        row[2] = cy * sx

        row = m[2]
        row[0] = sz * sx + cz * sy * cx
        row[1] = -cz * sx + sz * sy * cx
        row[2] = cy * cx

        return m

    # ----------------------------------------------------------------------
    def __add__(self, B):
        """Add 2 matrices"""
        if self.rows != B.rows or self.cols != B.cols:
            raise Exception("Matrix.add: matrices same size")
        m = Matrix(self.rows, self.cols)
        for i in range(self.rows):
            mrow = m[i]
            arow = self[i]
            brow = B[i]
            for j in range(self.cols):
                mrow[j] = arow[j] + brow[j]
        return m

    # ----------------------------------------------------------------------
    def __sub__(self, B):
        """Subtract 2 matrices"""
        if self.rows != B.rows or self.cols != B.cols:
            raise Exception("Matrix.add: matrices same size")
        m = Matrix(self.rows, self.cols)
        for i in range(self.rows):
            mrow = m[i]
            arow = self[i]
            brow = B[i]
            for j in range(self.cols):
                mrow[j] = arow[j] - brow[j]
        return m

    # ----------------------------------------------------------------------
    def __neg__(self):
        """Negate matrix"""
        m = Matrix(self.rows, self.cols)
        for i in range(self.rows):
            mrow = m[i]
            mold = self[i]
            for j in range(self.cols):
                mrow[j] = -mold[j]
        return m

    # ----------------------------------------------------------------------
    def __mul__(self, B):
        """Multiply two matrices or vector
           A.__mul__(B|vec) <==> A*B or A*vec"""
        if isinstance(B, Matrix):  # has to be a matrix of same cxN * Nxr
            if self.cols != B.rows:
                raise Exception("arrays don't have the correct dimensions")
            r = Matrix(self.rows, B.cols)
            for i in range(self.rows):
                for j in range(B.cols):
                    s = 0.0
                    for k in range(self.cols):
                        s += self[i][k] * B[k][j]
                    r[i][j] = s
            return r

        elif isinstance(B, list):  # Vector or list
            vecsize = len(B)
            v = Vector(vecsize)
            for i in range(vecsize):
                for j in range(min(self.cols, vecsize)):
                    v[i] += self[i][j] * B[j]
                for j in range(vecsize, self.cols):
                    v[i] += self[i][j]
            return v

        else:
            for row in self:
                for i in range(self.cols):
                    row[i] *= B
            return self

    # -----------------------------------------------------------------------
    # Special function to multiply a transformation matrix with a vector
    # ignoring the translation
    # -----------------------------------------------------------------------
    def multNoTranslation(self, B):
        """Multiply matrix with a vector ignoring the translation part"""
        if not isinstance(B, list):
            raise Exception("Invalid operation")
        vecsize = len(B)
        v = Vector(vecsize)
        for i in range(vecsize):
            for j in range(min(self.cols, vecsize)):
                v[i] += self[i][j] * B[j]
        return v

    # ----------------------------------------------------------------------
    def inv(self):
        """Inverse matrix in place"""

        if self.rows != self.cols:
            raise Exception("inverting a non square matrix")
        index = [0] * self.rows
        self.__ludcmp(index)
        y = Matrix(self.rows)
        for j in range(self.rows):
            col = [0.0] * self.rows
            col[j] = 1.0
            self.__lubksb(index, col)
            for i in range(self.rows):
                y[i][j] = col[i]
        for j in range(self.rows):
            self[j] = y[j]

    inverse = inv

    # ----------------------------------------------------------------------
    def clone(self):
        """Clone matrix"""
        m = Matrix(self.rows, self.cols)
        for i in range(self.rows):
            m[i] = self[i][:]
        return m

    # ----------------------------------------------------------------------
    # determinant with Gauss method
    # ----------------------------------------------------------------------
    def det(self, eps=_accuracy):
        """determinant of square matrix using Gauss method"""
        if self.rows == 2:
            return self[0][0] * self[1][1] - self[1][0] * self[0][1]
        elif self.rows == 3:
            return self[0][0] * (self[1][1] * self[2][2] - self[2][1] * self[1][2]) \
                - self[0][1] * (self[1][0] * self[2][2] - self[2][0] * self[1][2]) \
                + self[0][2] * (self[1][0] * self[2][1] - self[2][0] * self[1][1])

        M = self.clone()
        s = 1.0
        n = M.rows
        for i in range(n - 1):
            # find the absolute maximum value
            ma = abs(M[i][i])
            k = i
            for j in range(i + 1, n):
                if abs(M[j][i]) > ma:
                    ma = abs(M[j][i])
                    k = j
            if ma < eps:
                return 0.0

            # swap rows i,k
            if i != k:
                s = -s  # Change sign of determinate
                for j in range(n):
                    d = M[i][j]
                    M[i][j] = M[k][j]
                    M[k][j] = d

            # make all the following rows with zero at the i column
            for j in range(i + 1, n):
                if abs(M[j][i]) < _accuracy:
                    continue
                d = -M[i][i] / M[j][i]
                s *= d
                for k in range(i, n):
                    M[j][k] = M[i][k] + d * M[j][k]

        d = M[0][0] / s
        for i in range(1, n):
            d *= M[i][i]
        return d

    determinant = det

    # ----------------------------------------------------------------------
    # LU decomposition.
    # Parameters
    #      index[0:size]	row permutation record
    # ----------------------------------------------------------------------
    def __ludcmp(self, index):  # procedure expose indx.
        size = self.rows
        vv = [0.0] * size
        for i in range(size):
            big = 0.0
            for j in range(size):
                big = max(abs(self[i][j]), big)
            if big == 0:
                raise Exception("Singular matrix found")
            vv[i] = 1.0 / big

        for j in range(size):
            for i in range(j):
                s = self[i][j]
                for k in range(i):
                    s -= self[i][k] * self[k][j]
                self[i][j] = s

            big = 0.0
            for i in range(j, size):
                s = self[i][j]
                for k in range(j):
                    s -= self[i][k] * self[k][j]

                self[i][j] = s
                dum = vv[i] * abs(s)
                if dum >= big:
                    big = dum
                    imax = i

            if j != imax:
                for k in range(size):
                    dum = self[imax][k]
                    self[imax][k] = self[j][k]
                    self[j][k] = dum
                vv[imax] = vv[j]

            index[j] = imax
            if self[j][j] == 0.0:
                self[j][j] = 1E-20

            if j != size - 1:
                dum = 1.0 / self[j][j]
                for i in range(j + 1, size):
                    self[i][j] *= dum

    # ----------------------------------------------------------------------
    # backward substitution
    #      index[0:size]	  row permutation record
    #      col[0:size]	  right hand vector (?)
    # ----------------------------------------------------------------------
    def __lubksb(self, index, col):
        ii = -1
        size = self.rows
        for i in range(size):
            ip = index[i]
            s = col[ip]
            col[ip] = col[i]
            if ii >= 0:
                for j in range(ii, i):
                    s -= self[i][j] * col[j]
            elif s != 0.0:
                ii = i
            col[i] = s

        for i in range(size - 1, -1, -1):
            s = col[i]
            for j in range(i + 1, size):
                s -= self[i][j] * col[j]
            col[i] = s / self[i][i]


# -------------------------------------------------------------------------------
# Basic Matrices
# -------------------------------------------------------------------------------
Matrix.O = Matrix(4, type=0)  # noqa: E741
Matrix.U = Matrix(4, type=1)


# -------------------------------------------------------------------------------
# Quaternion
#
# Note: See the following for more information on quaternions:
#
# - Shoemake, K., Animating rotation with quaternion curves, Computer
#   Graphics 19, No 3 (Proc. SIGGRAPH'85), 245-254, 1985.
# - Pletinckx, D., Quaternion calculus as a basic tool in computer
#   graphics, The Visual Computer 5, 2-13, 1989.
# -------------------------------------------------------------------------------
class Quaternion(list):
    def __init__(self, a, b=None, c=None, d=None):
        if isinstance(a, Quaternion):
            self.extend(a)

        elif isinstance(a, Matrix):
            tr = a[0][0] + a[1][1] + a[2][2] + 1.0  # trace of matrix
            if tr > 0:
                S = math.sqrt(tr) * 2.0  # S=4*qw
                qw = 0.25 * S
                qx = (a[2][1] - a[1][2]) / S
                qy = (a[0][2] - a[2][0]) / S
                qz = (a[1][0] - a[0][1]) / S
            elif a[0][0] > a[1][1] and a[0][0] > a[2][2]:
                # S=4*qx
                S = math.sqrt(1.0 + a[0][0] - a[1][1] - a[2][2]) * 2.0
                qx = 0.25 * S
                qy = (a[0][1] + a[1][0]) / S
                qz = (a[0][2] + a[2][0]) / S
                qw = (a[2][1] - a[1][2]) / S
            elif a[1][1] > a[2][2]:
                # S=4*qy
                S = math.sqrt(1.0 + a[1][1] - a[0][0] - a[2][2]) * 2.0
                qx = (a[0][1] + a[1][0]) / S
                qy = 0.25 * S
                qz = (a[1][2] + a[2][1]) / S
                qw = (a[0][2] - a[2][0]) / S
            else:
                # S=4*qz
                S = math.sqrt(1.0 + a[2][2] - a[0][0] - a[1][1]) * 2.0
                qx = (a[0][2] + a[2][0]) / S
                qy = (a[1][2] + a[2][1]) / S
                qz = 0.25 * S
                qw = (a[1][0] - a[0][1]) / S
            self.extend([qx, qy, qz, qw])

        elif isinstance(a, Vector) and isinstance(b, float):
            s = math.sin(b / 2.0) / a.length()
            self.append(a[0] * s)
            self.append(a[1] * s)
            self.append(a[2] * s)
            self.append(math.cos(b / 2.0))

        else:
            self.extend([a, b, c, d])

    # ----------------------------------------------------------------------
    # Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0
    # If they don't add up to 1.0, dividing by their magnitued will
    # renormalize them.
    # ----------------------------------------------------------------------
    def norm(self):
        """normalize quaternion"""
        mag = math.sqrt(self[0]**2 + self[1]**2 + self[2]**2 + self[3]**2)
        self[0] /= mag
        self[1] /= mag
        self[2] /= mag
        self[3] /= mag
        return mag

    normalize = norm

    # ----------------------------------------------------------------------
    def vector(self):
        """return vector of quaternion"""
        return Vector(self[0], self[1], self[2])

    # ----------------------------------------------------------------------
    # return rotation matrix
    # ----------------------------------------------------------------------
    def matrix(self):
        """return rotation matrix"""
        m = Matrix(4, type=1)
        m[0][0] = 1.0 - 2.0 * (self[1] * self[1] + self[2] * self[2])
        m[0][1] = 2.0 * (self[0] * self[1] - self[2] * self[3])
        m[0][2] = 2.0 * (self[2] * self[0] + self[1] * self[3])

        m[1][0] = 2.0 * (self[0] * self[1] + self[2] * self[3])
        m[1][1] = 1.0 - 2.0 * (self[2] * self[2] + self[0] * self[0])
        m[1][2] = 2.0 * (self[1] * self[2] - self[0] * self[3])

        m[2][0] = 2.0 * (self[2] * self[0] - self[1] * self[3])
        m[2][1] = 2.0 * (self[1] * self[2] + self[0] * self[3])
        m[2][2] = 1.0 - 2.0 * (self[1] * self[1] + self[0] * self[0])
        return m

    # ----------------------------------------------------------------------
    # Given two rotations, e1 and e2, expressed as quaternion rotations,
    # figure out the equivalent single rotation and stuff it into dest.
    # This routine also normalizes the result every RENORMCOUNT times it is
    # called, to keep error from creeping in.
    # ----------------------------------------------------------------------
    def __add__(self, b):
        v1 = self.vector()
        v2 = b.vector()
        t1 = v1 * b[3]
        t2 = v2 * self[3]
        t3 = v2.cross(v1)
        tf = t1 + t2 + t3
        q = Quaternion(tf, self[3] * b[3] - v1.dot(v2))
        q.norm()
        return q

    # ----------------------------------------------------------------------
    def __iadd__(self, b):
        v1 = self.vector()
        v2 = b.vector()
        t1 = v1 * b[3]
        t2 = v2 * self[3]
        t3 = v2.cross(v1)

        tf = t1 + t2 + t3

        self[0] = tf[0]
        self[1] = tf[1]
        self[2] = tf[2]
        self[3] = self[3] * b[3] - v1.dot(v2)
        self.norm()
        return self


# -------------------------------------------------------------------------------
def gauss(A, B):
    """Solve A*X = B using the Gauss elimination method"""

    n = len(A)
    s = [0.0] * n
    X = [0.0] * n

    p = list(range(n))
    for i in range(n):
        s[i] = max([abs(x) for x in A[i]])

    for k in range(n - 1):
        # select j>=k so that
        # |A[p[j]][k]| / s[p[i]] >= |A[p[i]][k]| / s[p[i]] for i = k,k+1,...,n
        j = k
        ap = abs(A[p[j]][k]) / s[p[j]]
        for i in range(k + 1, n):
            api = abs(A[p[i]][k]) / s[p[i]]
            if api > ap:
                j = i
                ap = api

        if j != k:
            p[k], p[j] = p[j], p[k]  # Swap values

        for i in range(k + 1, n):
            z = A[p[i]][k] / A[p[k]][k]
            A[p[i]][k] = z
            for j in range(k + 1, n):
                A[p[i]][j] -= z * A[p[k]][j]

    for k in range(n - 1):
        for i in range(k + 1, n):
            B[p[i]] -= A[p[i]][k] * B[p[k]]

    for i in range(n - 1, -1, -1):
        X[i] = B[p[i]]
        for j in range(i + 1, n):
            X[i] -= A[p[i]][j] * X[j]
        X[i] /= A[p[i]][i]

    return X


# -------------------------------------------------------------------------------
def solveOverDetermined(A, B, W=None):
    """Solve the overdetermined linear system defined by the matrices A,B
        such as A*X = B
    Optionally a weight can be specified"""
    if A.rows < A.cols:
        raise Exception("solveOverDetermined: " "A matrix has more columns than rows")
    AT = A.transpose()
    if W:
        Wd = Matrix.diagonal(W)
        ATA = AT * Wd * A
        ATB = AT * Wd * B
    else:
        ATA = AT * A
        ATB = AT * B
    ATA.inv()
    RT = ATA * ATB
    return [RT[i][0] for i in range(len(RT))]


# -------------------------------------------------------------------------------
def linear(X, Y):
    """
    Solve linear regression y = ax + b
    @return a,b,r
    """
    Sx = Sy = Sx2 = Sy2 = Sxy = 0.0
    for x, y in zip(X, Y):
        Sx += x
        Sy += y
        Sx2 += x * x
        Sy2 += y * y
        Sxy += x * y

    n = float(len(X))
    try:
        b = (Sxy - Sx * Sy / n) / (Sx2 - Sx * Sx / n)
        a = Sy / n - b * Sx / n
        r = (Sxy - Sx * Sy / n) / math.sqrt(Sx2 - Sx * Sx / n) * \
            math.sqrt(Sy2 - Sy * Sy / n)
        return a, b, r

    except ZeroDivisionError:
        return None


# -------------------------------------------------------------------------------
#   Idiotimes pragmatikwv symmetrikwv pivakwv
#
#   O algori8mos poy xrnsimopoieitai stnv roytiva eivai gvwstos sav
#   proseggistikn me8odos Jacobi.
#   O algori8mos ekmetaleyetai tnv idiotnta poy exoyv oi diagwvioi
#   pivakes, dnladn pivakes me mndevika ola ta stoixeia ektos tns
#   kyrias diagwvioy, va exoyv sav idiotimes ta diagwvia stoixeia.
#   Me tov metasxnmatismo.
#         T                       T
#      A1 = R1 (f) A R1(f),    A2 = R2 (f) A1 R2(f)
#   metaballoyme syvexws tov pivaka A, mexris otoy to a8roisma olwv
#   twv mn diagwviwv stoixeiwv f8asei mia ka8orismevn timn tns eklogns
#   toy xrnstn n givei mndev
#   Ta bnmata tns diadikasias eivai:
#   1. Avazntnsn toy apolytws megistoy mn diagwvioy stoixeioy
#      Divei ta p kai q
#   2. Prosdiorismos tns gwvias peristrofns f. Divei ta sinf kai cosf
#   3. Metasxnmatismos Ai -> Ai+1
#   4. Elegxos av to a8roisma twv mn diagwviwv stoixeiwv exei f8asei tnv
#      epi8ymntn timn. Eav vai tote ta diagwvia stoixeia eivai oi
#      proseggiseis twv idiotimwv, eav oxi tote epistrefoyme sto 1.
#      px.     |  1 -2 -1 |
#          A = | -2  1 -1 |
#              | -1 -1 2.5|
#      apolyto megisto A(1,2) = -2
#      Ypologizoyme tnv gwvia f, co=cos(f), si=sin(f) kai kavoyme tov
#      metasxnmatismo
#          | co -si  0 |   |  1 -2 -1 |   |  co  si  0 |
#      A = | si  co  0 | x | -2  1 -1 | x | -si  co  0 |
#          |  0   0  1 |   | -1 -1 2.5|   |   0   0  1 |
#
#
#      Oi parametroi tns roytivas eivai oi e3ns:
#    A     - pivakas tetragwvikos
#    eps   - akribeia (a8roisma tetragwvwv)
#    check - av prepei va elejei tnv symmetria toy arxikoy pivaka
#        n oxi
# -------------------------------------------------------------------------------
def eigenvalues(M, eps=_accuracy, check=False):
    """Return eigen values and eigen vectors of a symmetric matrix"""
    n = M.rows

    # elegxos av eivai symmetrikos o pivakas
    if check:
        if n != M.cols:
            return None
        for i in range(n):
            for j in range(i, n):
                if M[i][j] != M[j][i]:
                    return None

    # Allocate arrays
    A = M.clone()
    R = Matrix(n, type=0)
    RT = Matrix(n, type=0)
    ZW = Matrix(n, type=0)
    V = None

    # kavovika 8a prepei meta apo merikes prospa8eies va tov aporiptei
    while True:
        # Bnma 1. Avazntnsn toy apolytws megistoy mn diagwvioy stoixeioy
        p = 0
        q = 1
        el = abs(A[p][q])
        for i in range(1, n):
            for j in range(i):
                if abs(A[i][j]) > el:
                    el = abs(A[i][j])
                    p = i
                    q = j
            if el == 0:
                break

        # Ftiaxvei ta R, RT
        for i in range(n):
            for j in range(n):
                R[i][j] = RT[i][j] = (i == j)

        # Bnma 2. Prosdiorizei tnv gwvia f, cosf kai sinf
        fi = (A[q][q] - A[p][p]) / (2.0 * A[p][q])
        t = 1.0 / (fi + math.sqrt(fi * fi + 1.0))
        if fi < 0:
            t = -t
        co = 1.0 / math.sqrt(1.0 + t * t)
        si = t / math.sqrt(1.0 + t * t)

        R[p][p] = R[q][q] = co
        RT[p][p] = RT[q][q] = co

        R[p][q] = si
        R[q][p] = -si
        RT[p][q] = -si
        RT[q][p] = si

        # Bnma 3. metasxnmatismos Ai+1 = Rt * Ai * R
        #  ka8os kai to ginomeno Rn*...*R2*R1 that
        #  gives us the eigenvectors
        if V is None:
            V = R.clone()
        else:
            V = V * R

        for i in range(n):
            for j in range(n):
                if j not in (p, q):
                    ZW[i][j] = A[i][j]
                else:
                    zw1 = 0
                    for k in range(n):
                        zw1 += A[i][k] * R[k][j]
                    ZW[i][j] = zw1

        for i in range(n):
            for j in range(n):
                if i not in (p, q):
                    A[i][j] = ZW[i][j]
                else:
                    zw1 = 0
                    for k in range(n):
                        zw1 += RT[i][k] * ZW[k][j]
                    A[i][j] = zw1

        # Bnma 4. Briskoymai to a8roisma kai elegxoyme av teleiwse
        zw1 = 0
        k = 0
        for i in range(1, n):
            for j in range(i):
                zw1 += A[i][j] * A[i][j]
            k += 1
        zw1 /= n

        # Exit condition
        if zw1 <= eps:
            break
    return ([A[i][i] for i in range(n)], V.T())


# -------------------------------------------------------------------------------
# Given a function, and given a bracketing triplet of abscissas ax,bx,cx (such
# that bx is between ax and cx, and f(bx) is less than both f(ax) and f(cx),
# this routing performs a golden section search for the minimum, isolating it
# to a fractional precision of about eps. The abscissa of the minimum is
# returned as xmin, and the minimum function value is returned as golden, the
# returned function value.
#
# @param func	function to be evaluated
# @param ax	triplet of abscissas ax,bx,cx
# @param bx	where func(x+bx*d) < min[ func(x+ax*d), func(x+cx*d) ]
# @param cx	...
# @param x	starting vector/value
# @param d	direction vector/value
# @param eps	accuracy of search
# -------------------------------------------------------------------------------
def goldenSectionSearch(func, ax, bx, cx, x, d=1, eps=_accuracy):
    R = 0.61803399  # The golden ratio
    C = (1.0 - R)
    x0 = ax  # At any given time we will keep track of four points
    x3 = cx  # x0, x1, x2, x3
    if abs(cx - bx) > abs(bx - ax):
        x1 = bx
        x2 = bx + C * (cx - bx)
    else:
        x2 = bx
        x1 = bx - C * (bx - ax)

    f1 = func(x + x1 * d)  # The initial function evaluation
    f2 = func(x + x2 * d)
    while abs(x3 - x0) > eps * (abs(x1) + abs(x2)):
        if f2 < f1:
            x0 = x1
            x1 = x2
            x2 = R * x1 + C * x3
            f1 = f2
            f2 = func(x + x2 * d)
        else:
            x3 = x2
            x2 = x1
            x1 = R * x2 + C * x0
            f2 = f1
            f1 = func(x + x1 * d)

    if f1 < f2:
        return x1
    else:
        return x2


# -------------------------------------------------------------------------------
# Generators for calculating a) the permutations of a sequence and
# b) the combinations and selections of a number of elements from a
# sequence. Uses Python 2.2 generators.
# Similar solutions found also in comp.lang.python
# Keywords: generator, combination, permutation, selection
#
# See also: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/105962
# See also: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66463
# See also: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66465
# -------------------------------------------------------------------------------
def xcombinations(items, n):
    if n <= 0:
        yield []
    else:
        for i in range(len(items)):
            for cc in xcombinations(items[:i] + items[i + 1:], n - 1):
                yield [items[i]] + cc


def xuniqueCombinations(items, n):
    if n <= 0:
        yield []
    else:
        for i in range(len(items)):
            for cc in xuniqueCombinations(items[i + 1:], n - 1):
                yield [items[i]] + cc


def xselections(items, n):
    if n <= 0:
        yield []
    else:
        for i in range(len(items)):
            for ss in xselections(items, n - 1):
                yield [items[i]] + ss


def xpermutations(items):
    return xcombinations(items, len(items))


# -------------------------------------------------------------------------------
# Conversion between rectangular and polar coordinates
# Usage:
#  real, real = rect(real, real [, deg=False])
#  real, real = polar(real, real [, deg=False])
# Normally, rect() and polar() uses radian for angle; but,
# if deg=True specified, degree is used instead.
# -------------------------------------------------------------------------------
# radian if deg=False; degree if deg=True
def rect(r, w, deg=False):
    """
    Convert from polar (r,w) to rectangular (x,y)
        x = r cos(w)
        y = r sin(w)
    """
    if deg:
        w = math.radians(w)
    return r * math.cos(w), r * math.sin(w)


# -------------------------------------------------------------------------------
# radian if deg=False; degree if deg=True
# -------------------------------------------------------------------------------
def polar(x, y, deg=False):
    """
    Convert from rectangular (x,y) to polar (r,w)
        r = sqrt(x^2 + y^2)
        w = arctan(y/x) = [-pi,pi] = [-180,180]
    """
    if deg:
        return math.hypot(x, y), math.degrees(math.atan2(y, x))
    else:
        return math.hypot(x, y), math.atan2(y, x)


def quadratic(b, c, eps=_accuracy):
    D = b * b - 4.0 * c
    if D <= 0.0:
        x1 = -0.5 * b  # Always return this as a solution!!!
        if D >= -eps * (b * b + abs(c)):
            return x1, x1
        else:
            return None, None
    else:
        if b > 0.0:
            bD = -b - math.sqrt(D)
        else:
            bD = -b + math.sqrt(D)
        return 0.5 * bD, 2.0 * c / bD


# -------------------------------------------------------------------------------
# Cubic equation: y^3 + a*y^2 + b*y + c = 0 (or ax^3 + bx^2 + cx + d = 0)
#
# Normally, x^3 + ax^2 + bx + c = 0 is assumed with the 3 coefficients as
# arguments; but, if 4 arguments are present, then ax^3 + bx^2 + cx + d = 0 is
# assumed.
#
# Even though both quadratic() and cubic() functions take real arguments, they
# can be modified to accept any real or complex coefficients because the method
# of solution does not make any assumptions.
# -------------------------------------------------------------------------------
def cubic(a, b, c, d=None, eps=_accuracy):
    if d is not None:  # (ax^3 + bx^2 + cx + d = 0)
        a, b, c = b / float(a), c / float(a), d / float(a)

    Q = (a * a - 3.0 * b) / 9.0
    R = (2. * a**3 - 9. * a * b + 27. * c) / 54.

    R2 = R**2
    Q3 = Q**3
    if R2 < Q3:  # the cubic has 3 real solutions
        theta = math.acos(R / math.sqrt(Q3))
        sqrt_Q = math.sqrt(Q)
        x1 = -2. * sqrt_Q * math.cos(theta / 3.) - a / 3.
        x2 = -2. * sqrt_Q * math.cos((theta + 2. * math.pi) / 3.) - a / 3.
        x3 = -2. * sqrt_Q * math.cos((theta - 2. * math.pi) / 3.) - a / 3.
        return x1, x2, x3

    A = -math.copysign(1.0, R) * (abs(R) + math.sqrt(R2 - Q3))**(1. / 3.)
    if abs(A) > eps:
        B = Q / A
    else:
        B = 0.0

    return (A + B) - a / 3., None, None

# imaginary roots
# x2 = -(A+B)/2 - a/3 + i*sqrt(3)*(A-B)
# x3 = -(A+B)/2 - a/3 - i*sqrt(3)*(A-B)


# -------------------------------------------------------------------------------
# Fit a plane to a set of points using least square fitting
# -------------------------------------------------------------------------------
def fitPlane(xyz):
    # First do statistics with points
    Sx = Sy = Sz = 0.0
    Sx2 = Sy2 = Sz2 = 0.0
    Sxy = Syz = Sxz = 0.0
    for x, y, z in xyz:
        Sx += x
        Sy += y
        Sz += z

        Sx2 += x**2
        Sy2 += y**2
        Sz2 += z**2

        Sxy += x * y
        Syz += y * z
        Sxz += x * z

    n = float(len(xyz))
    Sx /= n
    Sy /= n
    Sz /= n
    Vx = Sx2 / n - Sx**2
    Vy = Sy2 / n - Sy**2
    Vz = Sz2 / n - Sz**2

    # Count zero variances
    nv = int(abs(Vx) <= _accuracy) + int(abs(Vy) <= _accuracy) + \
        int(abs(Vz) <= _accuracy)
    if nv > 1:
        return None
    elif nv == 1:
        # Planes parallel to axes
        # Try the solution of x=Xo or y=Yo or z=Zo
        if abs(Vx) <= _accuracy:
            return 1.0, 0.0, 0.0, -Sx
        elif abs(Vy) <= _accuracy:
            return 0.0, 1.0, 0.0, -Sy
        else:
            return 0.0, 0.0, 1.0, -Sz

    # Try a generic solution
    #  z = ax + by + d    <=>  ax + by -z + d = 0
    #  assuming c=-1
    #  it can only fail on ax + by + d = 0
    #
    #  / Sx2    Sxy    Sx \       / Sxz \
    #  | Sxy    Sy2    Sy | * X = | Syz |
    #  \ Sx     Sy     n  /       \ Sz  /
    A = Matrix([[Sx2, Sxy, Sx], [Sxy, Sy2, Sy], [Sx, Sy, n]])
    B = Matrix([[Sxz], [Syz], [Sz]])

    try:
        A.inverse()
        X = A * B
        return X[0][0], X[1][0], -1.0, X[2][0]
    except Exception:
        pass

    # Try a solution where c=0
    # y = ax + d  <=>   ax -y +d = 0
    # .
    #  / Sx2    Sx \       / Sxy \
    #  |           | * X = |     |
    #  \ Sx     n  /       \ Sy  /
    A = Matrix([[Sx2, Sx], [Sx, n]])
    B = Matrix([[Sxy], [Sy]])
    try:
        A.inverse()
        X = A * B
        return X[0][0], -1.0, 0.0, X[1][0]
    except Exception:
        return None


# -------------------------------------------------------------------------------
# Evaluating n'th degree polynomial is simple loop, starting with highest
# coefficient a[n].
# -------------------------------------------------------------------------------
def polyeval(a, x):
    """
    p(x) = polyeval(a, x)
         = a[0] + a[1]x + a[2]x^2 +...+ a[n-1]x^{n-1} + a[n]x^n
         = a[0] + x(a[1] + x(a[2] +...+ x(a[n-1] + a[n]x)...)
    """
    p = 0
    a.reverse()
    for coef in a:
        p = p * x + coef
    a.reverse()
    return p


# -------------------------------------------------------------------------------
# Find the first derivative of a polynomial
# -------------------------------------------------------------------------------
def polyderiv(a):
    """
    p'(x) = polyderiv(a)
          = b[0] + b[1]x + b[2]x^2 +...+ b[n-2]x^{n-2} + b[n-1]x^{n-1}
    where b[i] = (i+1)a[i+1]
    """
    b = []
    for i in range(1, len(a)):
        b.append(i * a[i])
    return b


# -------------------------------------------------------------------------------
# Factor out a root from n'th degree polynomial, and return the remaining
# (n-1)'th degree polynomial.
# list = polyreduce(list, number)
# -------------------------------------------------------------------------------
def polyreduce(a, root):
    """
    Given x = r is a root of n'th degree polynomial p(x) = (x-r)q(x),
    divide p(x) by linear factor (x-r) using the same algorithm as
    polynomial evaluation. Then, return the (n-1)'th degree quotient
    q(x) = polyreduce(a, r)
         = c[0] + c[1]x + c[2]x^2 +...+ c[n-2]x^{n-2} + c[n-1]x^{n-1}
    """
    c, p = [], 0
    a.reverse()
    for coef in a:
        p = p * root + coef
        c.append(p)
    a.reverse()
    c.reverse()
    return c[1:]


# -------------------------------------------------------------------------------
# Conversion from integer to Roman
# -------------------------------------------------------------------------------
def int2roman(num):
    """
    Convert an integer to Roman numeral
    """
    if not isinstance(num, int):
        raise TypeError("expected integer, got %s" % type(input))

    if not 0 < num < 4000:
        raise ValueError("Argument must be between 1 and 3999")

    ints = (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1)
    nums = ('M', 'CM', 'D', 'CD', 'C', 'XC', 'L', 'XL', 'X', 'IX', 'V', 'IV', 'I')
    result = ""
    for i in range(len(ints)):
        count = int(num / ints[i])
        result += nums[i] * count
        num -= ints[i] * count
    return result


# -------------------------------------------------------------------------------
# Conversion from Roman to integer
# -------------------------------------------------------------------------------
def roman2int(roman):
    """
    convert a roman string to integer
    """
    if not isinstance(roman, str):
        raise TypeError("expected string, got %s" % type(roman))
    roman = roman.upper()
    nums = ('M', 'D', 'C', 'L', 'X', 'V', 'I')
    ints = (1000, 500, 100, 50, 10, 5, 1)
    places = []
    for c in roman:
        if c not in nums:
            raise ValueError("input is not a valid roman numeral: %s" % roman)
    for i in range(len(roman)):
        c = roman[i]
        value = ints[nums.index(c)]
        # If the next place holds a larger number, this value is negative.
        try:
            nextvalue = ints[nums.index(roman[i + 1])]
            if nextvalue > value:
                value *= -1
        except IndexError:
            # there is no next place.
            pass
        places.append(value)
    sum = 0
    for n in places:
        sum += n

    # Easiest test for validity...
    if int2roman(sum) == roman:
        return sum
    else:
        raise ValueError('input is not a valid roman numeral: %s' % roman)


# ===============================================================================
if __name__ == "__main__":
    from pymchelper.flair.common.log import say

    for i in range(4, 50):
        num = "-1e-%d" % (i)
        say(num, repr(float(num)), format_number(num, 10))

    digits = "1234567890123456789012345678"
    for i in range(len(digits)):
        num = digits[:i]
        say("%02d %40s |%10s|" % (i, num, format_number(num, 10)))
    say()
    for i in range(len(digits)):
        num = digits[:i] + ".123"
        say("%02d %40s |%10s|" % (i, num, format_number(num, 10)))
    say()
    for i in range(len(digits)):
        num = digits[:i] + ".1234567890"
        say("%02d %40s |%10s|" % (i, num, format_number(num, 10)))
    say()
    for i in range(len(digits)):
        num = "-" + digits[:i]
        say("%02d %40s |%10s|" % (i, num, format_number(num, 10)))
    say()
    for i in range(len(digits)):
        num = "-" + digits[:i] + ".123"
        say("%02d %40s |%10s|" % (i, num, format_number(num, 10)))
    say()
    for i in range(len(digits)):
        num = "-" + digits[:i] + ".1234567890"
        say("%02d %40s |%10s|" % (i, num, format_number(num, 10)))

    say(Vector.O)
    a = 1.2345678901234567890
    w = -1e-05
    say(format_number(w, 22))
    say(format_number(a, 22))
    say(Vector.Y.direction(0.0000001))
    a = Vector(range(5))
    a.norm()
    say("a=", a, "len(a)=", a.length())
    b = Vector(range(5))
    say("b=", b, "length(b)=", b.length(), "len(b)=", len(b))

    say("matO\n", Matrix.O)
    say("matU\n", Matrix.U)

    M = Matrix(4)
    a = Vector(0.1, 0.2, 0.3)
    M.rotate(math.radians(45.0), a)
    say("M\n", M)
    say("a=", a)
    say("M*a=", M * a)
    say("M*X=", M * Vector.X)

    say(d2s(dms(1, 2, 3.006)))

    M = Matrix([[1.0, -2.0, -3.0, 1.0], [-2.0, 1.0, -2.0, -1.0], [-3.0, -2.0, 2.0, 2.0], [1.0, -1.0, 2.0, 3.0]])
    say("M=\n", M)
    say("det(M)=", M.det())
    v, u = eigenvalues(M)
    say("Eigenvalues=", v)
    say("Eigenvectors=\n", u)
    say("uT * diag(v) * u=\n", u.T() * Matrix.diagonal(v) * u)

    say()
    say("Permutations of 'love'")
    for p in xpermutations(['l', 'o', 'v', 'e']):
        say(''.join(p))

    say()
    say("Combinations of 2 letters from 'love'")
    for c in xcombinations(['l', 'o', 'v', 'e'], 2):
        say(''.join(c))

    say()
    say("Unique Combinations of 2 letters from 'love'")
    for uc in xuniqueCombinations(['l', 'o', 'v', 'e'], 2):
        say(''.join(uc))

    say()
    say("Selections of 2 letters from 'love'")
    for s in xselections(['l', 'o', 'v', 'e'], 2):
        say(''.join(s))

    say()
    say(map(''.join, list(xpermutations('done'))))

    say()
    say("cubic(x^3-3x^2-10x+24=0)=", cubic(-3.0, -10.0, 24.0))

    say()
    for i in range(1, 51):
        say(i, int2roman(i))



from enum import IntEnum


class SHDetType(IntEnum):
    """
    List of available detector types below is based on IDET(5,*) in detect.f and sh_scoredef.h in SHIELD-HIT12A.
    If new detectors are added, class SHEstimator in estimator.py should also be updated.
    """

    none = 0
    energy = 1
    fluence = 2
    crossflu = 3
    letflu = 4

    dose = 5
    dlet = 6
    tlet = 7
    avg_energy = 8
    avg_beta = 9

    spc = 10
    material = 11
    ddd = 12
    alanine = 13
    counter = 14

    pet = 15
    dletg = 16
    tletg = 17
    zone = 18
    medium = 19

    rho = 20
    q = 21
    flu_char = 22
    flu_neut = 23
    flu_neqv = 24

    angle = 25
    trace = 26
    kinetic_energy = 27
    energy_nuc = 28
    energy_amu = 29

    a = 30
    amass = 31
    amu = 32
    gen = 33
    id = 34

    dedx = 35
    mass_dedx = 36
    track_length = 37
    nkerma = 38
    dose_gy = 39

    dose_eqv = 40
    eqv_dose = 41
    user1 = 42
    user2 = 43
    n_eqv_dose = 44  # Neutron equivalent dose, ICRP 103 protection quantity.

    z2beta2 = 45
    dose_av_z2beta2 = 46
    track_av_z2beta2 = 47
    dose_av_q = 48
    track_av_q = 49

    z = 50
    zeff = 51
    zeff2beta2 = 52
    tzeff2beta2 = 53
    dzeff2beta2 = 54

    count = 55  # simple counter, not normalized to per primary particle
    norm_count_point = 56  # score_point counter, normalized per primary particle
    count_point = 57  # score_point counter, not normalized per primary particle
    moca_yf = 58  # frequency-averaged lineal energy from Moca
    moca_yd = 59  # dose-averaged lineal energy from Moca

    q_eff = 60  # Effective Q, Q_eff, aka zeff2beta2
    dq_eff = 61  # Dose-averaged Q_eff
    tq_eff = 62  # Track-averaged Q_eff

    mcpl = 63  # MC particle list, aka. phase space list

    let_bdo2016 = 120  # for differential scoring
    angle_bdo2016 = 121  # for differential scoring
    dose_gy_bdo2016 = 205
    alanine_gy_bdo2016 = 213

    invalid = 32767

    def __str__(self):
        return self.name.upper().replace('_', '-')
